# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mock
import unittest


class TestBsbWorker(unittest.TestCase):

    def setUp(self):
        super(TestBsbWorker, self).setUp()
        # settings.py creates file during import that requires unit_config
        # we must mock unit config before we can import this module
        m = mock.patch('ci_utils.unit_config.get')
        self.addCleanup(m.stop)
        m.start()
        from bsbuilder import run_worker
        self.worker = run_worker.BSBuilderWorker()

    fake_request_params = {
        'ticket_id': 1,
        'subtickets': [],
        'series': 'a_series',
        'ppa': 'a_ppa',
        'archive': 'an_archive',
        'progress_trigger': 'A_TRIGGER'
    }

    fake_subticket_status = [{
        'name': 'the_source',
        'version': '1.0',
        'status_text': 'A_STATUS',
        'message': 'SUB_MESSAGE',
    }]

    @mock.patch('bsbuilder.run_worker.watch_ppa.watch_ppa')
    @mock.patch('bsbuilder.run_worker.watch_ppa.check_ppa')
    @mock.patch('ci_utils.amqp_utils.progress_completed')
    @mock.patch('ci_utils.amqp_worker.AMQPWorker._create_data_store')
    def test_handle_request(self, mocked_create_data_store,
                            mocked_progress_completed,
                            mocked_check_ppa, mocked_watch_ppa):
        # `BSBuilderWorker.handle_request` uses watch_ppa helpers
        # to build all sources related to a ticket.
        mocked_log = mock.MagicMock()
        mocked_ds = mock.MagicMock()
        mocked_ds.put_file.return_value = 'swift_ref'
        mocked_create_data_store.return_value = mocked_ds
        # Creating successful watch_ppa mocks.
        mocked_check_ppa.return_value = (False, self.fake_subticket_status)
        mocked_watch_ppa.return_value = (0, self.fake_subticket_status)

        # Process a testing BSB request.
        cb, result = self.worker.handle_request(
            self.fake_request_params, mocked_log)
        # 'progress_complete' is returned at the end of the processing
        # with context information to the lander and the log messages
        # stored in Swift.
        self.assertEqual(mocked_progress_completed, cb)
        self.assertEquals(
            {'artifacts': [{'type': 'LOGS',
                            'name': 'package_build.output.log',
                            'reference': 'swift_ref'}],
             'message': 'The package build was successful.',
             'subticket_status': [{'message': 'SUB_MESSAGE',
                                   'version': '1.0',
                                   'name': 'the_source',
                                   'status_text': 'A_STATUS'}],
             'ppa': 'a_ppa',
             'archive': 'an_archive'},
            result)
        # A ticket data_store acessor is created.
        mocked_create_data_store.assert_called_once_with(
            self.fake_request_params['ticket_id'])
        # log.info() is called for announcing the package was
        # successfully built.
        mocked_log.info.assert_called_with(
            'The package build was successful.')

    @mock.patch('bsbuilder.run_worker.watch_ppa.check_ppa')
    @mock.patch('ci_utils.amqp_utils.progress_failed')
    @mock.patch('ci_utils.amqp_worker.AMQPWorker._create_data_store')
    def test_handle_request_check_failure(
            self, mocked_create_data_store, mocked_progress_failed,
            mocked_check_ppa):
        # `BSBuilderWorker.handle_request` failures store enough
        # tracing information.
        mocked_log = mock.MagicMock()
        mocked_ds = mock.MagicMock()
        mocked_ds.put_file.return_value = 'swift_ref'
        mocked_create_data_store.return_value = mocked_ds
        # Creating successful watch_ppa mocks.
        mocked_check_ppa.return_value = (True, self.fake_subticket_status)

        # Process a testing BSB request.
        cb, result = self.worker.handle_request(
            self.fake_request_params, mocked_log)

        # 'progress_failed' is called informing the lander of the issue.
        self.assertEqual(mocked_progress_failed, cb)
        self.assertEqual(
            {'artifacts': [{'type': 'LOGS',
                            'name': 'package_build.output.log',
                            'reference': 'swift_ref'}],
             'message': 'Source package upload failed.',
             'subticket_status': [{'message': 'SUB_MESSAGE',
                                   'version': '1.0',
                                   'name': 'the_source',
                                   'status_text': 'A_STATUS'}],
             'ppa': 'a_ppa',
             'archive': 'an_archive'},
            result)

        # log.error() is called once stating the problem.
        mocked_log.error.assert_called_once_with(
            'Source package upload failed.')
