# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import os
import requests
import urlparse

from ci_cli import utils
from ci_utils import ticket_states


class ImageObjectNotFound(Exception):
    """Raised when image object cannot be found."""


def _download(ticket, dest_path):
    """Download the ticket IMAGE artifact to the given path.

    Raises ImageObjectNotFound() in case of error.
    """
    # Lookup a suitable IMAGE artifact for the given ticket.
    query = '?type__exact=IMAGE&ticket__exact={}'.format(ticket)
    url = urlparse.urljoin(
        utils.CI_URL, utils.TICKET_ARTIFACT_BASE + query)
    images = utils.get(url)
    if len(images['objects']) == 0:
        raise ImageObjectNotFound(
            'No images found for ticket number {}.'.format(ticket))
    image_name = images['objects'][0]['name']
    # Fetch the list of artifacts TempUrls for the ticket.
    url = urlparse.urljoin(
        utils.CI_URL, os.path.join(utils.GK_URL, 'ticket', str(ticket)))
    artifacts_urls = utils.get(url)
    temp_url = artifacts_urls.get(image_name)
    if temp_url is None:
        raise ImageObjectNotFound('No tempurl for {}'.format(image_name))
    # Download the image artifact and write it to the destination path.
    response = requests.get(temp_url, stream=True)
    if response.status_code != 200:
        raise ImageObjectNotFound('Failed to download {}'.format(temp_url))
    # Create destination path parts if needed.
    if not os.path.exists(os.path.dirname(dest_path)):
        os.makedirs(os.path.dirname(dest_path))
    with open(dest_path, 'wb') as fp:
        for chunk in response.iter_content(chunk_size=4096):
            # Ignore keep-alives.
            if not chunk:
                continue
            fp.write(chunk)
            fp.flush()


def get_image(args):
    """Get IMAGE artifact.

    Find and download the image artifact for the given ticket (args.ticket)
    or the latest completed one.

    The image content will be stored on the specified path (args.name).
    """
    log = logging.getLogger('ubuntu-ci')

    if os.path.exists(args.name):
        log.error("{} already exists.".format(args.name))
        return

    ticket = args.ticket
    if not ticket:
        ticket_status_base = utils.CI_URL + utils.TICKET_STATUS_BASE
        url = ticket_status_base + '?current_workflow_step={}'.format(
            ticket_states.TicketWorkflowStep.COMPLETED.value)
        data = utils.get(url)
        if len(data['objects']) == 0:
            log.error("No completed tickets found.")
            return
        ticket = max([x['id'] for x in data.get('objects', [])])

    _download(ticket, args.name)

    if args.ticket:
        log.info(
            'Successfully downloaded image for ticket #{}.'.format(ticket))
    else:
        log.info('Successfully downloaded latest release candidate image.')

    log.info("Image file can be found at: {}".format(args.name))
