#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import ci_utils
import httplib
import json
import logging
import mock
import socket
from testfixtures import LogCapture
import urlparse

from ci_cli import utils
from ci_cli.tests import (
    capture_stderr,
    capture_stdout,
    MainScriptTestCase,
    FakeResponse,
)
from ci_utils.testing import get_test_file_path
from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


class CLICommandsTestCase(MainScriptTestCase):
    """Tests cli basic commands."""

    def setUp(self):
        super(CLICommandsTestCase, self).setUp()
        ci_utils.SUPPORTED_SERIES = ['trusty']

    @mock.patch('requests.get')
    def test_ticket_status_one_ticket(self, mock_requests):
        """Verify cli status command when providing ticket number."""
        data = {
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "My cool feature",
            "id": '5',
            "description": "my new feature",
            "owner": "user@example.com",
            "created": "Wed, 12 Mar 2014 02:18:28 +0000",
            "updated": "Wed, 12 Mar 2014 02:30:22 +0000",
            "bug_id": "12345",
            "added_binaries": "mynewpackage",
            "removed_binaries": "myoldpackage",
        }
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = ['status', '-t', '5']
        with capture_stdout(self.cli.main, args) as cm:
            self.assertEquals(
                cm,
                'Ticket #5 - My cool feature\n'
                'Curent landing step: {} - {}\n'
                'Description: my new feature\n'
                'Owner: user@example.com\n'
                'Created: Wed, 12 Mar 2014 02:18:28 +0000\n'
                'Updated: Wed, 12 Mar 2014 02:30:22 +0000\n'
                'Bug #: 12345\n'
                'Added binaries: mynewpackage\n'
                'Removed binaries: myoldpackage\n'.format(
                    str(TicketWorkflowStep.IMAGE_BUILDING),
                    str(TicketWorkflowStepStatus.INPROGRESS)))
        mock_requests.assert_called_once()

    @mock.patch('requests.get')
    def test_ticket_status_all_tickets(self, mock_requests):
        """Verify cli status command for multiple tickets."""
        data = {
            "objects": [
                {
                    "id": 5,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "My cool feature",
                },
                {
                    "id": 6,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "Ubuntu bug fix",
                },
            ]
        }
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = ['status']
        with capture_stdout(self.cli.main, args) as cm:
            self.assertEquals(
                cm,
                "Ticket #5, My cool feature is 100 110\n" +
                "Ticket #6, Ubuntu bug fix is 100 110\n")
        mock_requests.assert_called_once()

    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    @mock.patch('ci_cli.utils.patch')
    def test_create_ticket_multiple_subtickets(self, mock_patch, mock_post,
                                               mock_process):
        """Verify cli create_ticket command for multiple .changes.

        This also tests the case two sources are targeted to the same series.
        """
        changes_1 = get_test_file_path('foobar_0.1-1_source.changes')
        changes_2 = get_test_file_path('barfoo_0.1-1_source.changes')

        args = ['create_ticket', '-t', '"New feature"', '-b', '4321', '-o',
                'someone@example.com', '-d', '"This is a cool new feature"',
                '-a', 'foo,bar', '-r', 'baz', '-s', changes_1, '-s', changes_2]

        with mock.patch('sys.stdout'):
            logger = logging.getLogger()
            self.cli.main(args, log=logger)

        self.assertEquals(mock_process.call_count, 2)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)

    def test_create_ticket_multiple_subtickets_different_suites(self):
        """Verify cli create_ticket command for multiple .changes."""
        # We don't really need to support these, just need 2 series
        ci_utils.SUPPORTED_SERIES = ['trusty', 'raring']
        changes_1 = get_test_file_path('barfoo_0.1-1_source.changes')
        changes_2 = get_test_file_path(
            'foobar-unsupported_0.1-1ubuntu1_source.changes')

        args = ['create_ticket', '-t', '"New feature"', '-b', '4321', '-o',
                'someone@example.com', '-d', '"This is a cool new feature"',
                '-a', 'foo,bar', '-r', 'baz', '-s', changes_1, '-s', changes_2]

        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertEquals(
            lc.records[0].msg,
            "All submitted packages have to target the same suite. You "
            "submitted packages with suites: trusty, raring")

    @mock.patch('requests.get')
    def test_ts_server_returns_urlerror(self, mock_requests):
        """Test cli response to ts server refusing connection (111)."""
        mock_requests.side_effect = [
            FakeResponse(
                400, '', socket.error(111, "Connection refused"))
        ]
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_requests.assert_called_once()
        self.assertTrue("Cannot reach the server at %s." % utils.CI_URL in
                        lc.records[0].msg, lc.records[0].msg)
        self.assertTrue("Connection refused" in lc.records[0].msg,
                        lc.records[0].msg)

    @mock.patch('requests.get')
    def test_ts_server_returns_internal_server_error(self, mock_requests):
        """Test cli response to ts server returning error 500."""
        mock_requests.side_effect = [
            FakeResponse(
                500, "", "INTERNAL SERVER ERROR"
            )
        ]
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_requests.assert_called_once()
        self.assertEquals("Server at {} reported internal error. Maybe a "
                          "database hiccup?".format(utils.CI_URL),
                          lc.records[0].msg)

    @mock.patch('requests.get')
    def test_ts_server_returns_url_not_found(self, mock_requests):
        """Test cli response to ts server returning 404 error."""
        mock_requests.side_effect = [
            FakeResponse(404, '', 'Not Found')
        ]
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
            mock_requests.assert_called_once()
            self.assertEquals("Server at {} not found. Is 'ci_url' correctly "
                              "set? (404 Client Error: Not Found)".format(
                                  utils.CI_URL), lc.records[0].msg)

    @mock.patch('requests.get')
    def test_ts_server_returns_bad_status_line(self, mock_requests):
        """Test cli response to ts server returning BadStatusLine."""
        args = ['status']
        mock_requests.side_effect = httplib.BadStatusLine('')
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_requests.assert_called_once()
        self.assertEquals("Server at {} replied with an empty response. Is "
                          "'ci_url' pointing to the correct service?".format(
                              utils.CI_URL), lc.records[0].msg)

    @mock.patch('requests.get')
    def test_ts_server_returns_other_httperror(self, mock_requests):
        """Test cli response to ts server returning HTTPError."""
        mock_requests.side_effect = [
            FakeResponse(403, '', 'Forbidden')
        ]
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_requests.assert_called_once()
        self.assertEquals(
            "Cannot reach the server at {}. Is the "
            "server reachable from this machine? "
            "(403 Client Error: Forbidden).".format(utils.CI_URL),
            lc.records[0].msg)

    @mock.patch('ci_cli.utils.post',
                return_value='http://mocked/api/v1/ticket/1/')
    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.utils.patch')
    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_create_ticket(self, mocked_post, mocked_put, mocked_of,
                           mocked_upa, mocked_ug, mocked_upo):
        # CLI 'create_ticket' action allows users to upload new sources
        # via gatekeeper (without swift credentials).
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(201, '/ticket.html?ticket_id=1'),
        ]
        mocked_put.side_effect = [
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
        ]
        mocked_of.return_value = 'a_content'
        foobar_changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['-v2', 'create_ticket', '-t', '"New feature"', '-d',
                '"New feature description"', '-o', 'someone@example.com',
                '-a', 'foobar', '-s', foobar_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        ticket_url = urlparse.urljoin(
            utils.CI_URL, '/ticket.html?ticket_id=1')
        # The first output line contains a local absolute host path, so
        # it is ignored.
        self.assertEquals(
            ['Parsing foobar_0.1-1_source.changes...',
             'Validating .changes...',
             'foobar_0.1-1 parsed and validated.',
             'Created source package upload: http://mocked/api/v1/ticket/1/',
             'Created subticket: http://mocked/api/v1/ticket/1/',
             'Uploading foobar_0.1-1.dsc ...',
             'Uploading foobar_0.1.orig.tar.gz ...',
             'Uploading foobar_0.1-1.debian.tar.gz ...',
             'Uploading foobar_0.1-1_source.changes ...',
             'Done! Access {} for more information.'.format(ticket_url),
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Created artifact: http://mocked/api/v1/ticket/1/'],
            list([r.msg for r in lc.records]))

    @mock.patch('ci_cli.utils.post',
                return_value='http://mocked/api/v1/ticket/1/')
    @mock.patch('ci_cli.utils.get')
    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    @mock.patch('requests.patch')
    def test_create_ticket_update_fails(
            self, mocked_patch, mocked_post, mocked_put, mocked_of, mocked_get,
            mocked_upost):
        # CLI 'create_ticket' action allows users to upload new sources
        # via gatekeeper (without swift credentials).
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(201, '/ticket.html?ticket_id=1'),
        ]
        mocked_put.side_effect = [
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
        ]
        mocked_of.return_value = 'a_content'
        foobar_changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['-v2', 'create_ticket', '-t', '"New feature"', '-d',
                '"New feature description"', '-o', 'someone@example.com',
                '-a', 'foobar', '-s', foobar_changes]
        mocked_patch.side_effect = [
            FakeResponse(401, {}, "Failed to patch ticket")]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        ticket_url = urlparse.urljoin(
            utils.CI_URL, '/ticket.html?ticket_id=1')
        # The first output line contains a local absolute host path, so
        # it is ignored.
        self.assertEquals(
            ['Parsing foobar_0.1-1_source.changes...',
             'Validating .changes...',
             'foobar_0.1-1 parsed and validated.',
             'Created source package upload: http://mocked/api/v1/ticket/1/',
             'Created subticket: http://mocked/api/v1/ticket/1/',
             'Uploading foobar_0.1-1.dsc ...',
             'Uploading foobar_0.1.orig.tar.gz ...',
             'Uploading foobar_0.1-1.debian.tar.gz ...',
             'Uploading foobar_0.1-1_source.changes ...',
             'Done! Access {} for more information.'.format(ticket_url),
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Created artifact: http://mocked/api/v1/ticket/1/',
             'Ticket was created successfully but there was a problem adding '
             'it to the queue: 401 Client Error: Failed to patch ticket. This '
             'should be a transient error, please contact an admin. Your '
             'ticket number is 1.'],
            list([r.msg for r in lc.records]))


class CliArgumentsTestCase(MainScriptTestCase):
    """Test cli arguments validation."""

    def _test_parser_missing_arguments(self, args, expected_msg):
        """Helper to capture error output when cli fails."""
        with capture_stderr(self.assertRaises, SystemExit, self.cli.main,
                            args) as cm:
            self.assertNotEqual(cm, None)
        self.assertTrue(expected_msg in cm)

    def test_cli_arguments_are_required(self):
        """Verify cli arguments are required."""
        args = ['create_ticket']
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -t/--title is required")

        args.extend(['-t', '"New feature"'])
        self._test_parser_missing_arguments(
            args,
            "create_ticket: error: argument -d/--description is required")

        args.extend(['-d', '"This is a cool new feature"'])
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -o/--owner is required")

        args.extend(['-o', 'someone@example.com'])
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -a/--add is required")

        args.extend(['-a', 'foo'])
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -s/--sources is required")

        # All arguments were provided. It should fail because file is not
        # found.
        args.extend(['-s', 'foobar.changesss'])
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertTrue("Changes file not found" in lc.records[0].msg)

    def test_cli_only_source_files_pass(self):
        """Verify cli fails in case file isn't prefixed .changes."""
        wrong_changes = get_test_file_path('foobar_0.1-1.dsc')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Upload file must be a .changes; you provided '{}'".format(
                wrong_changes))

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_is_valid_package_list_add_space(self, mock_patch):
        """Test a ticket adding test binaries and not removing any fails."""
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo, bar', '-s', changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg[0],
            "Added and removed packages have to be comma-separated lists "
            "with no spaces.")

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_is_valid_package_list_remove_space(self, mock_patch):
        """Test a ticket removing two test binaries fails."""
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-r', 'bar, baz', '-s', changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg[0],
            "Added and removed packages have to be comma-separated lists "
            "with no spaces.")


class ChangesValidationTestCase(MainScriptTestCase):
    """Test how cli handles malformed, unreleased or unsigned .changes."""

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_dsc_file_in_disguise(self, mock_patch):
        """Verify cli behavior when .changes file is a .dsc."""
        wrong_changes = get_test_file_path('foobar_0.1-1_dsc.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem parsing .changes file, are you sure it's a valid "
            ".changes? (Maybe this is a .dsc file?)")

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_changes_file_is_something_else(self, mock_patch):
        """Verify cli behavior when provided .changes file isn't a .changes."""
        wrong_changes = get_test_file_path('foobar_0.1-1_random.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem parsing .changes file, are you sure it's a valid "
            ".changes? (Changes file could not be parsed.)")

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/40/')
    def test_cli_changes_file_unsigned_fails(self, mock_patch):
        """Verify cli behavior when provided .changes isn't signed."""
        unsigned_changes = get_test_file_path(
            'foobar_0.1-1_unsigned.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', unsigned_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            with mock.patch('sys.stdout'):
                self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem validating .changes file: failed to verify signature. Is "
            "it signed? (No signature on)")

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_changes_file_is_unreleased(self, mock_patch):
        """Verify cli behavior when provided .changes file is UNRELEASED."""
        changes_filename = 'foobar_0.1-1_unreleased.changes'
        wrong_changes = get_test_file_path(changes_filename)
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem validating .changes file: 'Distribution' 'UNRELEASED' in "
            "{} is not supported. Supported suites: {}".format(
                changes_filename,
                ", ".join(ci_utils.valid_suites())))

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_changes_file_targets_unsupported_series(self, mock_patch):
        """Verify cli behavior when .changes targets unsupported series."""
        changes_filename = 'foobar-unsupported_0.1-1ubuntu1_source.changes'
        wrong_changes = get_test_file_path(changes_filename)
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem validating .changes file: 'Distribution' 'raring' in {} "
            "is not supported. Supported suites: {}".format(
                changes_filename,
                ", ".join(ci_utils.valid_suites())))
