#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import json

from ci_cli.tests import (
    capture_stdout,
    MainScriptTestCase,
    FakeResponse,
)
from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


class GetTicketStatus(MainScriptTestCase):

    @mock.patch('requests.get')
    def test_get_single_ticket_status(self, mock_requests):
        data = {
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "My cool feature",
            "id": '100',
            "description": "my new feature",
            "owner": "user@example.com",
            "created": "Wed, 12 Mar 2014 02:18:28 +0000",
            "updated": "Wed, 12 Mar 2014 02:30:22 +0000",
            "bug_id": "12345",
            "added_binaries": "mynewpackage",
            "removed_binaries": "myoldpackage",
        }
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = self.cli.parse_arguments(['status', '-t', '100'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                'Ticket #100 - My cool feature\n'
                'Curent landing step: {} - {}\n'
                'Description: my new feature\n'
                'Owner: user@example.com\n'
                'Created: Wed, 12 Mar 2014 02:18:28 +0000\n'
                'Updated: Wed, 12 Mar 2014 02:30:22 +0000\n'
                'Bug #: 12345\n'
                'Added binaries: mynewpackage\n'
                'Removed binaries: myoldpackage\n'.format(
                    str(TicketWorkflowStep.IMAGE_BUILDING),
                    str(TicketWorkflowStepStatus.INPROGRESS)),
                cm)
        mock_requests.assert_called_once()

    @mock.patch('requests.get')
    def test_get_ticket_status_unicode(self, mock_requests):
        data = {
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.INPROGRESS),
            "title": "Testing unicode in title: joão",
            "id": '100',
            "description": "my new feature",
            "owner": "user@example.com",
            "created": "Wed, 12 Mar 2014 02:18:28 +0000",
            "updated": "Wed, 12 Mar 2014 02:30:22 +0000",
            "bug_id": "12345",
            "added_binaries": "mynewpackage",
            "removed_binaries": "myoldpackage",
        }
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = self.cli.parse_arguments(['status', '-t', '100'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                'Ticket #100 - Testing unicode in title: joão\n'
                'Curent landing step: {} - {}\n'
                'Description: my new feature\n'
                'Owner: user@example.com\n'
                'Created: Wed, 12 Mar 2014 02:18:28 +0000\n'
                'Updated: Wed, 12 Mar 2014 02:30:22 +0000\n'
                'Bug #: 12345\n'
                'Added binaries: mynewpackage\n'
                'Removed binaries: myoldpackage\n'.format(
                    str(TicketWorkflowStep.IMAGE_BUILDING),
                    str(TicketWorkflowStepStatus.INPROGRESS)),
                cm)
        mock_requests.assert_called_once()

    @mock.patch('requests.get')
    def test_get_single_failed_ticket_status(self, mock_requests):
        data = {
            "current_workflow_step": str(TicketWorkflowStep.FAILED),
            "status": str(TicketWorkflowStepStatus.FAILED),
            "title": "My cool feature",
            "id": '100',
            "description": "my new feature",
            "owner": "user@example.com",
            "created": "Wed, 12 Mar 2014 02:18:28 +0000",
            "updated": "Wed, 12 Mar 2014 02:30:22 +0000",
            "bug_id": "12345",
            "added_binaries": "mynewpackage",
            "removed_binaries": "myoldpackage",
        }
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = self.cli.parse_arguments(['status', '-t', '100'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                'Ticket #100 - My cool feature\n'
                '{}\n'
                'Description: my new feature\n'
                'Owner: user@example.com\n'
                'Created: Wed, 12 Mar 2014 02:18:28 +0000\n'
                'Updated: Wed, 12 Mar 2014 02:30:22 +0000\n'
                'Bug #: 12345\n'
                'Added binaries: mynewpackage\n'
                'Removed binaries: myoldpackage\n'.format(
                    str(TicketWorkflowStepStatus.FAILED)),
                cm)
        mock_requests.assert_called_once()

    @mock.patch('requests.get')
    def test_get_all_ticket_status(self, mock_requests):
        data = {
            "objects": [
                {
                    "id": 5,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "My cool feature",
                },
                {
                    "id": 6,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "Ubuntu bug fix",
                },
                {
                    "id": 7,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "Upstream bug fix",
                },
                {
                    "id": 9,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "Other new feature",
                },
            ]
        }
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(data))
        ]
        args = self.cli.parse_arguments(['status'])
        with capture_stdout(args.func, args) as cm:
            self.assertEqual(
                "Ticket #5, My cool feature is 100 110\n" +
                "Ticket #6, Ubuntu bug fix is 100 110\n" +
                "Ticket #7, Upstream bug fix is 100 110\n" +
                "Ticket #9, Other new feature is 100 110\n",
                cm)
        mock_requests.assert_called_once()

    @mock.patch('requests.get')
    def test_get_status_404_response(self, mock_requests):
        mock_requests.side_effect = [
            FakeResponse(404, '', 'Not Found')
        ]
        args = self.cli.parse_arguments(['status', '-t', '99'])
        with self.assertRaises(SystemExit) as cm:
            args.func(args)
        mock_requests.assert_called_once()
        self.assertEqual("Ticket number 99 not found.", cm.exception.message)
