#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import mock
import os
from testfixtures import LogCapture

from ci_cli import utils
from ci_cli.image import ImageObjectNotFound
from ci_cli.tests import (
    MainScriptTestCase,
    FakeResponse,
)


artifact_data = {
    "objects": [
        {
            "id": 6,
            "name": "d62d3d9a-ac3d-11e3-847d-fa163eaf5928.img",
            "reference": "http://www.example.com",
            "resource_uri": "/api/v1/ticketartifact/6/",
            "ticket": {
                "id": 4,
            },
            "type": "IMAGE"
        },
    ]
}

ticket_data = {
    "objects": [
        {
            "id": 1,
        },
        {
            "id": 4,
        },
        {
            "id": 2,
        },
    ],
}

artifact_tempurls = {
    'd62d3d9a-ac3d-11e3-847d-fa163eaf5928.img': 'http://A_TEMP_URL',
}


class TestGetTicketImage(MainScriptTestCase):

    def setUp(self):
        """Create an image file for tests."""
        super(TestGetTicketImage, self).setUp()
        self.image_path = os.path.join(
            utils.HOME, 'd62d3d9a-ac3d-11e3-847d-fa163eaf5928.img')

        def _remove_image(image_path):
            if os.path.exists(image_path):
                os.remove(image_path)

        self.addCleanup(_remove_image, self.image_path)

    @mock.patch('requests.get')
    def test_get_ticket_image(self, mock_requests):
        # An image is found via Ticket-System artifacts list, the
        # corresponding tempurl is obtained from Gatekeeper, then
        # the image file is downloaded and stored in the specified
        # path.
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(artifact_data)),
            FakeResponse(200, json.dumps(artifact_tempurls)),
            FakeResponse(200, 'IMAGE_CONTENT'),
        ]
        args = self.cli.parse_arguments(
            ['get_image', '-t', '4', '-n', self.image_path])
        with LogCapture() as lc:
            args.func(args)
        self.assertEqual('IMAGE_CONTENT', open(self.image_path).read())
        self.assertEquals(
            ['Successfully downloaded image for ticket #4.',
             'Image file can be found at: {}'.format(self.image_path)],
            [r.msg for r in lc.records])

    @mock.patch('requests.get')
    def test_get_ticket_image_not_found(self, mock_requests):
        # If the corresponding tempurl for a registered image cannot
        # be found ImageObjectNotFound() is raised.
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(artifact_data)),
            FakeResponse(200, json.dumps({})),
        ]
        args = self.cli.parse_arguments(
            ['get_image', '-t', '4', '-n', self.image_path])
        with self.assertRaises(ImageObjectNotFound) as cm:
            args.func(args)
        self.assertEquals(
            'No tempurl for {}'.format(os.path.basename(self.image_path)),
            cm.exception.message)

    def test_get_ticket_image_already_exists(self):
        # If the specified image destination path already exists, there
        # is not external requests and a message indicating the problem
        # is rendered.
        open(self.image_path, 'a').close()
        args = self.cli.parse_arguments(
            ['get_image', '-t', '4', '-n', self.image_path])
        with LogCapture() as lc:
            args.func(args)
        self.assertEquals(
            '{} already exists.'.format(self.image_path),
            lc.records[0].msg)

    @mock.patch('requests.get')
    def test_get_last_completed_ticket_image(self, mock_requests):
        # If a ticket number is not specified, CLI looks for the
        # latest COMPLETED ticket and downloads its image.
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps(ticket_data)),
            FakeResponse(200, json.dumps(artifact_data)),
            FakeResponse(200, json.dumps(artifact_tempurls)),
            FakeResponse(200, 'IMAGE_CONTENT'),
        ]
        args = self.cli.parse_arguments(
            ['get_image', '-n', self.image_path])
        args.func(args)
        self.assertEqual('IMAGE_CONTENT', open(self.image_path).read())

    @mock.patch('requests.get')
    def test_get_last_completed_ticket_none_complete(self, mock_requests):
        # If no ticket is specified and there are no COMPLETED tickets
        # the CLI renders a message indicating the problem.
        mock_requests.side_effect = [
            FakeResponse(200, json.dumps({'objects': []})),
        ]
        args = self.cli.parse_arguments(
            ['get_image', '-n', self.image_path])
        with LogCapture() as lc:
            args.func(args)
        self.assertEquals(
            'No completed tickets found.', lc.records[0].msg)
