#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import unittest

from ci_cli.tests import (
    FakeResponse,
    MainScriptTestCase,
)
from ci_cli.ticket import (
    new_ticket,
    SourcePackageUpload,
    SourcePackageUploadError,
    SubTicket,
    Ticket,
)
from ci_utils.testing import (
    TestCaseWithGnupg,
    get_test_file_path
)


class SourcePackageUploadTestCase(unittest.TestCase):
    """Tests for `SourcePackageUpload` context manager."""

    @mock.patch('requests.post')
    def test_create_fails(self, mocked_post):
        # `create` raises `SourcePackageUploadError` in case of error.
        spu = SourcePackageUpload('1', 'http://server/')
        self.assertIsNone(spu.sandbox_url)
        mocked_post.side_effect = [
            FakeResponse(404, 'Zoing!')
        ]
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.create()
        self.assertEquals(
            'Cannot create sandbox for ticket 1 '
            '(http://server/gk/api/v1/ticket/1/sandbox/ - 404)',
            cm.exception.message)
        self.assertIsNone(spu.sandbox_url)

    @mock.patch('requests.post')
    def test_create_works(self, mocked_post):
        # `create` creates a SPU sandbox in Gatekepper storing its path.
        spu = SourcePackageUpload('1', 'http://server/')
        self.assertIsNone(spu.sandbox_url)
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID')
        ]
        spu.create()
        self.assertEquals(
            'http://server/gk/ticket/1/sandbox/UUID', spu.sandbox_url)

    @mock.patch('requests.put')
    def test_upload_needs_init(self, mocked_put):
        # `upload` raises an error if the sandbox context is not initialised.
        spu = SourcePackageUpload('1')
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.upload('a_path')
        self.assertEquals('SPU not initialized.', cm.exception.message)
        mocked_put.assert_not_called()

    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_upload_fail(self, mocked_post, mocked_put, mocked_of):
        # `upload` raises an error if it cannot get a tempurl from GK
        # of cannot PUT on the swift tempurl.
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        mocked_put.side_effect = [
            FakeResponse(404, 'Not Found'),
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(401, 'Not Authorized'),
        ]
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.upload('not_found_filename')
        self.assertEquals(
            "Cannot get TempUrl for 'not_found_filename' "
            "(http://server/gk/ticket/1/sandbox/UUID/not_found_filename "
            "- 404)",
            cm.exception.message)
        mocked_of.return_value = 'whatever'
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.upload('a_path')
        self.assertEquals(
            "Cannot use TempUrl 'A_TEMP_URL'", cm.exception.message)

    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_upload_works(self, mocked_post, mocked_put, mocked_of):
        # `upload` returns the GK response content (created object
        # sandbox path) when it works.
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        mocked_put.side_effect = [
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
        ]
        mocked_of.return_value = 'a_content'
        result = spu.upload('a_file')
        self.assertEquals('/a_sandbox_object_url/', result)

    @mock.patch('requests.post')
    def test_process_fail(self, mocked_post):
        # `process` raises `SourcePackageUploadError` if the sandbox is
        # not initialised.
        spu = SourcePackageUpload('1', 'http://server/')
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.process()
        self.assertEquals('SPU not initialized.', cm.exception.message)
        mocked_post.assert_not_called()

    @mock.patch('requests.post')
    def test_process_error(self, mocked_post):
        # `process` raises `SourcePackageUploadError` with the GK
        # response content (upload remote validation error).
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(400, 'Problem!'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        with self.assertRaises(SourcePackageUploadError) as cm:
            spu.process()
        self.assertEquals(
            'SPU remote processing error: Problem!', cm.exception.message)

    @mock.patch('requests.post')
    def test_process_works(self, mocked_post):
        # `process` returns the GK response content (upload validation
        # results) when it works.
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(201, 'Success!'),
        ]
        spu = SourcePackageUpload('1', 'http://server/')
        spu.create()
        result = spu.process()
        self.assertEquals('Success!', result)

    @mock.patch('ci_cli.ticket._open_file')
    @mock.patch('requests.put')
    @mock.patch('requests.post')
    def test_context_manager(self, mocked_post, mocked_put, mocked_of):
        # Using `SourcePackageUpload` as a context manager makes your
        # life easier!
        mocked_post.side_effect = [
            FakeResponse(201, '/ticket/1/sandbox/UUID'),
            FakeResponse(201, 'Success!'),
        ]
        mocked_put.side_effect = [
            FakeResponse(200, 'A_TEMP_URL'),
            FakeResponse(201, '/a_sandbox_object_url/'),
        ]
        mocked_of.return_value = 'a_content'

        with SourcePackageUpload('1', 'http://server/') as spu:
            result = spu.upload('a_file')

        self.assertEquals('/a_sandbox_object_url/', result)
        self.assertEquals('Success!', spu.result)


class SubTicketTestCase(TestCaseWithGnupg):
    def setUp(self):
        super(SubTicketTestCase, self).setUp()
        self.changes = get_test_file_path('foobar_0.1-1_source.changes')

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/sourcepackage/4/')
    def test_create_sourcepackage(self, mock_post):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=4)
        with mock.patch('sys.stdout'):
            new_subticket._parse_changes(self.changes)
        sourcepackage_uri = new_subticket._create_sourcepackage()
        mock_post.assert_called_once()
        self.assertEqual(sourcepackage_uri, '/api/v1/sourcepackage/4/')

    @mock.patch('ci_cli.utils.get_sourcepackage_uri',
                return_value='/api/v1/sourcepackage/5/')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/spu/38/')
    def test_create_spu(self, mock_post, mock_sp_uri):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=3)
        with mock.patch('sys.stdout'):
            new_subticket._parse_changes(self.changes)
        new_subticket._create_spu()

    @mock.patch('ci_cli.utils.get_sourcepackage_uri', return_value='')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/sourcepackage/5/')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/spu/39/')
    def test_create_spu_sourcepackage_not_found(self, mock_spu_post, mock_post,
                                                mock_sp_uri):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=3)
        with mock.patch('sys.stdout'):
            new_subticket._parse_changes(self.changes)
        new_subticket._create_spu()

    @mock.patch(
        'ci_cli.utils.post',
        return_value='http://www.example.com/api/v1/subticketartifact/38/',
    )
    def test_create_artifact(self, mock_post):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=2)
        file = 'foobar_source.changes'
        location = 'http://www.example.com/path/to/foobar_source.changes'
        new_subticket._create_artifact(file, location)
        mock_post.assert_called_once()

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/subticket/1/')
    def test_create_subticket(self, mock_post):
        new_subticket = SubTicket(owner='foobar@example.com', ticket_id=1)
        new_subticket._create_subticket()


class TicketTestCase(MainScriptTestCase):

    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/38/')
    def test_create_ticket(self, mock_post):
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', 'New feature', '-b', '1234', '-o',
                'someone@example.com', '-d', 'This is a cool new feature',
                '-a', 'foo', '-r', 'baz', '-s', changes]

        new_ticket_ = Ticket()
        args = self.cli.parse_arguments(args)
        new_ticket_._create_ticket(args)

    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/38/')
    @mock.patch('ci_cli.utils.patch')
    def test_add_new_ticket(self, mock_patch, mock_post, mock_process):
        """Verify cli create_ticket command for one .changes."""
        changes = get_test_file_path('foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', 'New feature', '-b', '1234', '-o',
                'someone@example.com', '-d', 'This is a cool new feature',
                '-a', 'foo', '-r', 'bar,baz', '-s', changes]

        args = self.cli.parse_arguments(args)
        with mock.patch('sys.stdout'):
            new_ticket_ = new_ticket(args)
        self.assertEquals(mock_process.call_count, 1)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)
        self.assertEquals(len(new_ticket_.subtickets), 1)
        self.assertTrue("foobar" in new_ticket_.subtickets)

    @mock.patch('ci_cli.ticket.SubTicket._process')
    @mock.patch('ci_cli.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    @mock.patch('ci_cli.utils.patch')
    def test_add_new_ticket_multiple_subtickets(self, mock_patch, mock_post,
                                                mock_process):
        """Verify cli create_ticket command for multiple .changes.

        This also tests the case two sources are targeted to the same series.
        """
        changes_1 = get_test_file_path('foobar_0.1-1_source.changes')
        changes_2 = get_test_file_path('barfoo_0.1-1_source.changes')

        args = ['create_ticket', '-t', '"New feature"', '-b', '4321', '-o',
                'someone@example.com', '-d', '"This is a cool new feature"',
                '-a', 'foo,bar', '-r', 'baz', '-s', changes_1, '-s', changes_2]

        args = self.cli.parse_arguments(args)
        with mock.patch('sys.stdout'):
            new_ticket_ = new_ticket(args)
        self.assertEquals(mock_process.call_count, 2)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)

        self.assertEquals(len(new_ticket_.subtickets), 2)
        self.assertTrue("foobar" in new_ticket_.subtickets)
        self.assertTrue("barfoo" in new_ticket_.subtickets)
        self.assertEquals(u'trusty', new_ticket_.suite)
