# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import requests
import urlparse

from ci_cli import utils
from ci_utils.file_handler import ChangesProcessor
from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


log = logging.getLogger('ubuntu-ci')


def _open_file(filepath):
    """Local file opener for making tests easier."""
    return open(filepath)


class MismatchSuitesError(Exception):
    """Raised when submitted packages target multiple series."""


class UnsupportedSeriesError(Exception):
    """Raised when package series isn't supported."""


class UnsupportedSuiteError(Exception):
    """Raised when package series isn't supported."""


class SourcePackageUploadError(Exception):
    """Raised where there is an error during the SPU."""


class QueueingTicketError(Exception):
    """Raised when ticket failed to be set to Queued."""


class SourcePackageUpload(object):
    """Context manager helper for performing SPUs via Gatekeeper."""

    def __init__(self, ticket_id, ci_url=None):
        if ci_url is None:
            ci_url = utils.CI_URL
        self.gk_url = urlparse.urljoin(ci_url, '/gk')
        self.ticket_id = ticket_id
        self.sandbox_url = None
        self.result = None

    def __enter__(self):
        """Creates a remote sandbox context."""
        self.create()
        return self

    def __exit__(self, exc, value, tb):
        """Process the context sandbox."""
        self.result = self.process()

    def create(self):
        """Request a Gatekeeper Sandbox."""
        create_url = self.gk_url + '/api/v1/ticket/{}/sandbox/'.format(
            self.ticket_id)
        response = requests.post(create_url)
        if response.status_code != 201:
            raise SourcePackageUploadError(
                'Cannot create sandbox for ticket {} ({} - {})'.format(
                    self.ticket_id, create_url, response.status_code))
        sandbox_path = response.content
        self.sandbox_url = self.gk_url + sandbox_path

    def process(self):
        """Process the initialised Sandbox."""
        if self.sandbox_url is None:
            raise SourcePackageUploadError('SPU not initialized.')
        response = requests.post(self.sandbox_url)
        if response.status_code != 201:
            raise SourcePackageUploadError(
                'SPU remote processing error: {}'.format(response.content))
        return response.content

    def upload(self, filepath):
        """Upload the given file to the context Sandbox."""
        if self.sandbox_url is None:
            raise SourcePackageUploadError('SPU not initialized.')
        dest_url = self.sandbox_url + '/' + os.path.basename(filepath)
        response = requests.put(dest_url)
        if response.status_code != 200:
            filename = os.path.basename(filepath)
            raise SourcePackageUploadError(
                'Cannot get TempUrl for \'{}\' ({} - {})'.format(
                    filename, dest_url, response.status_code))
        temp_url = response.content
        response = requests.put(temp_url, data=_open_file(filepath))
        if response.status_code != 201:
            raise SourcePackageUploadError(
                'Cannot use TempUrl \'{}\''.format(temp_url))
        return response.content


def upload_files(ticket_id, files):
    """Upload files to a ticket container via Gatekeeper."""
    with SourcePackageUpload(ticket_id) as spu:
        for file_path in files:
            filename = os.path.basename(file_path)
            log.info('Uploading {} ...'.format(filename))
            try:
                spu.upload(file_path)
            except SourcePackageUploadError as exc:
                log.error(exc.message)
                raise
    ticket_url = urlparse.urljoin(utils.CI_URL, spu.result)
    log.info('Done! Access {} for more information.'.format(ticket_url))


class SubTicket():

    def __init__(self, owner, ticket_id=None):
        self.owner = owner
        self.ticket_id = ticket_id
        self.sourcepackage = ''
        self.files = ''
        self.version = ''

        self.spu_uri = ''
        self.subticket_uri = ''

    @property
    def ticket_uri(self):
        if self.ticket_id:
            return utils.TICKET_BASE + str(self.ticket_id) + '/'
        return ''

    def _parse_changes(self, changes_filepath):
        log.info("Parsing {}...".format(os.path.basename(changes_filepath)))
        changes = ChangesProcessor(
            changes_filepath=changes_filepath)
        log.info("Validating .changes...")
        self.files = changes.process()
        self.version = changes.source_version
        self.suite = changes.suite
        self.series = self.suite.split("-")[0]
        self.sourcepackage = changes.source_package_name
        self.version = changes.source_version
        log.info("{}_{} parsed and validated.".format(
            self.sourcepackage, self.version))

    def _process(self):
        self._create_spu()
        self._create_subticket()
        self._create_artifacts()

    def _create_artifacts(self):
        """Create artifacts for this ticket."""
        upload_files(self.ticket_id, self.files)
        for file_path in self.files:
            filename = os.path.basename(file_path)
            location = os.path.join(
                'ticket-{}'.format(self.ticket_id), filename)
            self._create_artifact(filename, location)

    def _create_sourcepackage(self):
        data = {
            "name": self.sourcepackage,
        }
        location = utils.post(
            utils.CI_URL + utils.SOURCEPACKAGE_BASE, data=data)
        sourcepackage_id = utils.parse_id(location)
        return utils.SOURCEPACKAGE_BASE + sourcepackage_id + '/'

    def _create_spu(self):
        sourcepackage_uri = utils.get_sourcepackage_uri(self.sourcepackage)
        if sourcepackage_uri == '':
            sourcepackage_uri = self._create_sourcepackage()
        data = {
            "sourcepackage": sourcepackage_uri,
            "version": self.version,
        }
        location = utils.post(utils.CI_URL + utils.SPU_BASE, data=data)
        self.spu_id = utils.parse_id(location)
        self.spu_uri = utils.SPU_BASE + self.spu_id + '/'
        log.info("Created source package upload: %s" % location)
        return location

    def _create_subticket(self):
        data = {
            "ticket": self.ticket_uri,
            "assignee": self.owner,
            "source_package_upload": self.spu_uri,
        }
        location = utils.post(utils.CI_URL + utils.SUBTICKET_BASE, data=data)
        self.subticket_id = utils.parse_id(location)
        self.subticket_uri = utils.SUBTICKET_BASE + self.subticket_id + '/'
        log.info("Created subticket: %s" % location)
        return location

    def _create_artifact(self, file, location):
        data = {
            "type": "SPU",
            "reference": location,
            "name": file,
            "subticket": self.subticket_uri,
        }
        location_ = utils.post(utils.CI_URL + utils.SUBTICKET_ARTIFACT_BASE,
                               data=data)
        log.info("Created artifact: %s" % location_)


class Ticket():

    def __init__(self):
        self.ticket_id = ''
        self.ticket_uri = ''
        self.uploads = []
        self.subtickets = {}
        self.suite = ''

    def _create_ticket(self, args):
        data = {
            "title": args.title,
            "owner": args.owner,
            "description": args.description,
            "series": self.suite.split("-")[0],
            "bug_id": args.bug,
            "added_binaries": args.add,
            "removed_binaries": args.remove,
        }
        location = utils.post(utils.CI_URL + utils.TICKET_BASE, data=data)
        self.ticket_id = utils.parse_id(location)
        self.ticket_uri = utils.TICKET_BASE + self.ticket_id + '/'
        return self.ticket_id

    def _update_ticket(self):
        """
        After a ticket has been created, it needs to be updated to have a
        status of QUEUED in order to be processed.
        """
        data = {
            "current_workflow_step": TicketWorkflowStep.QUEUED.value,
            "status": TicketWorkflowStepStatus.QUEUED.value,
        }
        urlbase = utils.CI_URL + utils.API_URL + 'updateticketstatus/'
        try:
            utils.patch(urlbase + self.ticket_id + '/', data=data)
            print("You have successfully submitted a ticket to the Ubuntu CI "
                  "Engine. Your ticket number is %s." % self.ticket_id)
        except Exception as exc:
            raise QueueingTicketError(
                "Ticket was created successfully but there was a problem "
                "adding it to the queue: {}. This should be a transient "
                "error, please contact an admin. Your ticket number is "
                "{}.".format(exc, self.ticket_id))

    def add_new_ticket(self, args):
        suites = []
        subtickets = []
        # First, process all sources and validate the series.
        for source in args.sources:
            subticket = SubTicket(owner=args.owner)
            subticket._parse_changes(source)
            suites.append(subticket.suite)
            subtickets.append(subticket)
        suite = set(suites)
        if len(suite) > 1:
            raise MismatchSuitesError(
                "All submitted packages have to target the same suite. "
                "You submitted packages with suites: {}".format(
                    ", ".join(suites)))
        # After the series/suite is validated, create the ticket and the
        # subtickets.
        self.suite = suite.pop()
        ticket_id = self._create_ticket(args)
        for subticket in subtickets:
            subticket.ticket_id = ticket_id
            subticket._process()
            self.subtickets[subticket.sourcepackage] = subticket
        self._update_ticket()


def new_ticket(args):
    new_ticket = Ticket()
    new_ticket.add_new_ticket(args)
    return new_ticket


def update_ticket(args):
    """Script action for updating a ticket."""
    ticket_id = args.id
    for source in args.sources:
        changes = ChangesProcessor(changes_filepath=source)
        files = changes.process()
        upload_files(ticket_id, files)
