#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import logging
import os
import sys
import time
import urllib2

import lander
import lander.workflow

from ci_utils import (
    amqp_utils,
    dump_stack,
    API_VERSIONS,
)


logger = lander.get_logger('lander_service_wrapper')
HERE = os.path.abspath(os.path.dirname(__file__))


class MessageSync(object):
    progress = logging.getLogger('PROGRESS_TRIGGER')

    def __init__(self, ticket, started_cb=None, completed_cb=None,
                 message_cb=None):
        if not started_cb:
            started_cb = lambda: True
        if not completed_cb:
            completed_cb = lambda x: True
        if not message_cb:
            message_cb = lambda x: True
        self.ticket = ticket
        self.started_cb = started_cb
        self.completed_cb = completed_cb
        self.message_cb = message_cb

    def on_message(self, msg):
        data = json.loads(msg.body)
        state = data['state']

        # print "state" as a label for the rest of the items in the list
        params = {x: y for (x, y) in data.iteritems() if x != 'state'}
        if params:
            self.progress.info('%s: %s', state, params)
        else:
            self.progress.info(state)

        self.message_cb(data)

        if state == 'STATUS':
            self.started_cb()
        elif state == 'COMPLETED':
            self.ticket.process_artifacts(data)
            if data['result'] == 'PASSED':
                self.completed_cb(True)
            else:
                self.completed_cb(False)

        msg.channel.basic_ack(msg.delivery_tag)
        if data.get('exit'):
            self.data = data
            # this will kick us out of the run-forever worker loop
            msg.channel.basic_cancel(msg.consumer_tag)


def _drain_progress(amqp_config, queue, ticket, started_cb=None,
                    completed_cb=None, message_cb=None):
    sync = MessageSync(ticket, started_cb, completed_cb, message_cb)
    amqp_utils.process_queue(amqp_config, queue, sync.on_message, delete=True)
    return sync.data


def _post(url, data):
    logger.info('Calling via POST [{}]: {}'.format(url, data))
    req = urllib2.Request(url)
    req.add_header('Content-Type', 'application/json')
    resp = urllib2.urlopen(req, json.dumps(data))
    if resp.getcode() == 204:
        return None  # 204 is an empty post reply
    return resp.headers['location']


def _patch(url, data):
    logger.info('Calling via PATCH [{}]: {}'.format(url, data))
    req = urllib2.Request(url)
    req.add_header('Content-Type', 'application/json')
    req.get_method = lambda: 'PATCH'
    urllib2.urlopen(req, json.dumps(data))


def _get(url):
    logger.info('Calling via GET [{}]'.format(url))
    resp = urllib2.urlopen(url)
    resp = resp.read()
    return json.loads(resp)


def _get_ppa(ppa_assigner_url, ticket_id):
    timeout = 15
    url = '%s/api/v1/ppa/' % ppa_assigner_url
    while True:
        try:
            loc = _post(url, {'ticket_id': ticket_id})
            return loc
        except urllib2.HTTPError as e:
            if e.getcode() != 404:
                raise e
            logging.warn('No free PPAs can be found. Retrying in %ds', timeout)
            time.sleep(timeout)


def _handle_ppa_assigner(amqp_config, workflow):
    loc = _get_ppa(workflow.ppa_url, workflow.ticket_id)
    data = _get(loc)
    result = {
        'result': 'PASSED',
        'state': 'COMPLETED',
        'ppa': data['name'],
        'location': loc,
    }
    return result


def _handle_ppa_assigner_release(amqp_config, workflow):
    try:
        step = workflow.get_step('ppa_assigner')
        url = step.get('location', False)
        if url:
            _patch(url, {'state': 'dirty'})
        else:
            logger.info('ppa was never reserved by this job')
    except KeyError:
        logger.info('ppa_assigner was not run. nothing to do')

    return {'result': 'PASSED'}


def _handle_bsbuilder(amqp_config, workflow):
    ticket = workflow.ticket_api()
    queue = '%s-bsbuilder' % workflow.progress_trigger

    request_parameters = ticket.get_full_ticket()
    params = {
        'ticket_id': workflow.ticket_id,
        'subtickets': request_parameters['subticket'],
        'series': request_parameters.get('series', ''),
        'ppa': workflow.get_step('ppa_assigner')['ppa'],
        'archive': request_parameters.get('master_ppa', ''),
        'progress_trigger': queue,
    }

    ticket.set_pkg_waiting()
    try:
        amqp_utils.send(amqp_utils.BSBUILDER_QUEUE, json.dumps(params), True)
        logger.info('starting progress handler...')
        return _drain_progress(amqp_config, queue, ticket,
                               ticket.set_pkg_inprogress,
                               ticket.set_pkg_complete,
                               ticket.set_subticket_progress)
    except:
        ticket.set_pkg_complete(False)
        raise


def _get_binary_packages(ticket, request_parameters):
    # Start with the set of golden binary packages
    package_list = set([x['name'] for x in ticket.get_binaries()])

    # Extract the list of added packages from the ticket request
    try:
        package_list = package_list | set(
            request_parameters.get('added_binaries', None).split(','))
    except AttributeError:
        # Expected when added_binaries is None.
        pass

    # Finally remove the list of removed binaries from the ticket request
    try:
        removed_list = set(
            request_parameters.get('removed_binaries', None).split(','))
        # Log a warning if any package in the removed_binaries is not in the
        # package_list. This is not an error because multiple queued tickets
        # may attempt to remove the same package. However, This could be an
        # indicator of a mis-typed package name.
        diff_list = removed_list - package_list
        if diff_list:
            logger.warning('Package(s) to remove not found: '
                           '{}'.format(list(diff_list)))
        package_list = package_list - removed_list

    except AttributeError:
        # Expected when removed_binaries is None.
        pass

    return list(package_list)


def _handle_image_builder(amqp_config, workflow):
    ticket = workflow.ticket_api()
    queue = '%s-imgbuilder' % workflow.progress_trigger

    # Extract the image parameters from the master request parameters
    request_parameters = ticket.get_full_ticket()

    params = {
        'ticket_id': workflow.ticket_id,
        'api_version': API_VERSIONS.get('image_builder'),
        'image': request_parameters['base_image'],
        'progress_trigger': queue,
    }
    ticket.set_image_waiting()
    try:
        amqp_utils.send(
            amqp_utils.IMAGE_BUILDER_QUEUE, json.dumps(params), True)
        logger.info('starting progress handler...')
        return _drain_progress(amqp_config, queue, ticket,
                               ticket.set_image_inprogress,
                               ticket.set_image_complete)
    except:
        ticket.set_image_complete(False)
        raise


def _handle_publisher(amqp_config, workflow):
    ticket = workflow.ticket_api()
    queue = '%s-publisher' % workflow.progress_trigger

    # XXX: Setting step to PKG_PUBLISHING here because for now this step
    # is executed only after image testing. When we start using publishing
    # between steps to check consistency, we'll need to verify if this is the
    # last step, so publisher will check consistency and then publish. One
    # idea is to introduce a step between image testing and publishing that
    # would signal publishing is ready to happen. -- Ursinha (2014-06-24)
    ticket.set_publishing_waiting()
    # Extract the image parameters from the master request parameters
    request_parameters = ticket.get_full_ticket()
    # XXX: If we miss any of the variables here, the error message will be
    # cryptic. Should we worry about malformed json? -- Ursinha (2014-05-30)
    params = {
        'ticket_id': workflow.ticket_id,
        'subtickets': request_parameters['subticket'],
        'series': request_parameters['series'],
        'source_archive': workflow.get_step('ppa_assigner')['ppa'],
        'destination_archive': request_parameters['master_ppa'],
        'current_workflow_step': request_parameters['current_workflow_step'],
        'progress_trigger': queue,
    }
    try:
        amqp_utils.send(
            amqp_utils.PUBLISHER_QUEUE, json.dumps(params), True)
        logger.info('starting progress handler...')
        return _drain_progress(amqp_config, queue, ticket,
                               ticket.set_publishing_inprogress,
                               ticket.set_publishing_complete)
    except:
        ticket.set_publishing_complete(False)
        raise


def _handle_test_runner(amqp_config, workflow):
    ticket = workflow.ticket_api()
    queue = '%s-testrunner' % workflow.progress_trigger
    request_parameters = ticket.get_full_ticket()
    # Packages can be available in the ppa used by the branch source builder as
    # well as in the ppa used by the publisher.
    ppa_list = [
        workflow.get_step('ppa_assigner')['ppa'],  # branch source builder
        request_parameters['master_ppa'],  # publisher
    ]
    # FIXME: Should we care about 'subticket', 'source_package_upload',
    # 'sourcepackage' not being defined ? -- vila 2014-07-21
    packages = []
    for subticket in request_parameters['subticket']:
        spu = subticket['source_package_upload']
        packages.append(spu['sourcepackage']['name'])

    params = {
        'ticket_id': workflow.ticket_id,
        'series': request_parameters.get('series', ''),
        'image_id': workflow.get_step('image_builder')['image_id'],
        'ppa_list': ppa_list,
        'package_list': packages,
        'progress_trigger': queue,
    }

    ticket.set_testing_waiting()
    try:
        amqp_utils.send(amqp_utils.TEST_RUNNER_QUEUE, json.dumps(params), True)
        logger.info('starting progress handler...')
        return _drain_progress(amqp_config, queue, ticket,
                               ticket.set_testing_inprogress,
                               ticket.set_testing_complete)
    except:
        ticket.set_testing_complete(False)
        raise


def _get_service_handlers():
    '''get the availble service handlers.

    This is a handy mechanism so we don't have to maintain a list of available
    service handlers. Any function prefixed with '_handle_' will be exposed to
    the user.
    '''
    return [x[8:] for x in globals() if x.startswith('_handle_')]


def _get_parser():
    parser = argparse.ArgumentParser(
        description='Wraps the REST API calls with a progress queue monitor.')
    parser.add_argument('--service',
                        required=True,
                        choices=_get_service_handlers(),
                        help='The service name to call.')
    parser.add_argument('--results-file',
                        required=True,
                        help='File containing ticket results')
    return parser


def main(args):
    amqp_config = amqp_utils.get_config()
    if not amqp_config:
        return 1  # the get_config code prints an error

    handler = globals()['_handle_' + args.service]
    logger.debug('Executing service wrapper: {}'.format(args.service))

    workflow = lander.workflow.load(args.results_file)

    rc = 1
    try:
        results = handler(amqp_config, workflow)
        if results['result'] == 'PASSED':
            rc = 0
    except urllib2.HTTPError as e:
        body = e.read()
        logging.exception('Unable to handle request: %s: %s', e, body)
        results = {
            'result': 'ERROR',
            'resp_code': e.reason,
            'resp_body': body,
        }
    except Exception as e:
        logging.exception('Unexpected error: %s', e)
        results = {
            'result': 'FAILED',
            'error': e.message,
        }

    with open(args.results_file, 'w') as f:
        workflow.set_results(args.service, results)
        workflow.save(f)

    return rc


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    args = _get_parser().parse_args()
    sys.exit(main(args))
