# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import http.server
import json
import os
import threading
import time
import urllib.error
import urllib.parse
import urllib.request


HERE = os.path.abspath(os.path.dirname(__file__))

# We have a few checks that take a while. The implementations for them use
# caching logic. This gives us a consistent place to manage this value.
CACHE_TIMEOUT = 60


class JSONStatus(object):
    def __init__(self):
        self.results = []

    def _add_result(self, name, value, status):
        self.results.append({
            'label': name,
            'value': value,
            'status': status,
        })

    def add_okay(self, name, value):
        self._add_result(name, value, 'okay')

    def add_fail(self, name, value):
        self._add_result(name, value, 'fail')

    def add_true_false(self, name, value, condition):
        '''adds an 'okay' result if condition is true, 'fail' otherwise'''
        if condition:
            self.add_okay(name, value)
        else:
            self.add_fail(name, value)

    def add_url_test(self, name, url):
        try:
            urllib.request.urlopen(url).read()
            self.add_okay(name, 'connected')
        except Exception as e:
            self.add_fail(name, str(e))

    def add_rabbit_configured(self):
        '''adds a result status for status of rabbit-configuration'''
        # we don't want to import this for all users of the API. Some
        # may not install rabbit (eg the ppa-assigner)
        from ci_utils import amqp_utils
        if amqp_utils.get_config() is not None:
            self.add_okay('rabbit configured', True)
        else:
            self.add_fail('rabbit configured', False)

    def add_rabbit_worker_health(self, queue):
        '''checks if there are any workers attached to the queue.'''
        from ci_utils import amqp_utils
        config = amqp_utils.get_config()
        if config is None:
            # we already report rabbit isn't configured, so no sense adding
            # another failure for this
            return

        mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        base = 'http://{}:55672'.format(config.AMQP_HOST)
        # NOTE: juju leaves this guest/guest and doesn't seem to expose
        # anything in the relation to override it
        mgr.add_password(None, base, 'guest', 'guest')
        handler = urllib.request.HTTPBasicAuthHandler(mgr)

        try:
            resp = urllib.request.build_opener(handler).open(
                base + '/api/queues'
            )
            content = resp.read()
            data = json.loads(content)
            consumers = None
            for q in data:
                if q['name'] == queue:
                    consumers = q['consumers']
                    break
            if consumers is None:
                self.add_fail('workers-online', 'no queue defined')
            else:
                self.add_true_false('workers-online', consumers, consumers > 0)
        except:
            self.add_fail('workers-online', 'unable to check')

    def add_data_store_health(self, ticket=1):
        from ci_utils import unit_config, data_store
        try:
            ds = data_store.create_for_ticket(
                ticket, unit_config.get_auth_config())
            ds.ensure_swift_client()
            self.add_okay('data-store', 'connected')
            return ds
        except Exception as e:
            self.add_fail('data-store', str(e))


class StatusHandler(http.server.BaseHTTPRequestHandler):
    # must be static because this object is create for each request
    _cache = (None, 0)

    def __init__(self, *args, **kwargs):
        http.server.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def _get_cached_response(self):
        query = urllib.parse.urlparse(self.path).query
        query = urllib.parse.parse_qs(query)

        resp, timeout = StatusHandler._cache
        if query.get('refresh', ['0'])[0] == '1' or time.time() > timeout:
            resp = json.dumps(self.health_cb().results)
            StatusHandler._cache = (resp, time.time() + CACHE_TIMEOUT)

        if 'jsonp' in query.get('format', []):
            cb = query.get('callback', ['cb'])[0]
            resp = cb + '(' + resp + ')'

        return resp.encode('utf-8')

    def do_GET(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/json')
        self.end_headers()
        resp = self._get_cached_response()
        self.wfile.write(resp)

    def log_message(self, format, *args):
        return


def run_server(health_cb, port=8080, threaded=True):
    httpd = http.server.HTTPServer(('0.0.0.0', port), StatusHandler)
    httpd.RequestHandlerClass.health_cb = health_cb
    if threaded:
        t = threading.Thread(target=httpd.serve_forever)
        t.daemon = True
        t.start()
        return httpd
    else:
        httpd.serve_forever()
