#!/usr/bin/env python3
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import logging
import re

from subunit import ByteStreamToStreamResult
from testtools import StreamToExtendedDecorator, TestResult


logging.basicConfig(level='INFO')
logger = logging.getLogger(__name__)


class FileSaveResult(TestResult):

    """A TestResult that saves a test summary, details, and attachments."""

    def __init__(self, data_store):
        super().__init__()
        self.data_store = data_store

        self.testcase_summaries = {}
        self.summary = dict(
            tests_run=0,
            errors=0,
            expected_failures=0,
            failures=0,
            num_attachments=0,
            run_time=0,
            skipped=0,
            unexpected_successes=0,
            tests=self.testcase_summaries
        )

        self.content_attachments = []
        self.testcase_details = dict(
            run_time=2.0,
            content_attachments=self.content_attachments
        )

    def addError(self, test, err=None, details=None):
        super().addError(test, err=err, details=details)
        self.summary['num_attachments'] += len(details)
        test.status = 'error'
        self.save_file_attachments(test, details)

    def addExpectedFailure(self, test, err=None, details=None):
        super().addExpectedFailure(test, err=err, details=details)
        self.summary['num_attachments'] += len(details)
        test.status = 'xfail'
        self.save_file_attachments(test, details)

    def addFailure(self, test, err=None, details=None):
        super().addFailure(test, err=err, details=details)
        self.summary['num_attachments'] += len(details)
        test.status = 'fail'
        self.save_file_attachments(test, details)

    def addSkip(self, test, reason=None, details=None):
        super().addSkip(test, reason=reason, details=details)
        self.summary['num_attachments'] += len(details)
        test.status = 'skip'
        self.save_file_attachments(test, details)

    def addSuccess(self, test, details=None):
        super().addSuccess(test, details=details)
        self.summary['num_attachments'] += len(details)
        test.status = 'pass'
        self.save_file_attachments(test, details)

    def addUnexpectedSuccess(self, test, details=None):
        super().addUnexpectedSuccess(test, details=details)
        self.summary['num_attachments'] += len(details)
        test.status = 'uxsuccess'
        self.save_file_attachments(test, details)

    def stopTest(self, test):
        super().stopTest(test)
        elapsed = test._timestamps[1] - test._timestamps[0]
        test_time = elapsed.total_seconds()
        self.summary['run_time'] += test_time
        self.testcase_summaries[test.id()] = [test.status, test_time]
        self.testcase_details['run_time'] = test_time
        self.testcase_details['status'] = test.status

    def stopTestRun(self):
        super().stopTestRun()
        self.summary['errors'] = len(self.errors)
        self.summary['expected_failures'] = len(self.expectedFailures)
        self.summary['failures'] = len(self.failures)
        self.summary['skipped'] = len(self.skip_reasons)
        self.summary['tests_run'] = self.testsRun
        self.summary['unexpected_successes'] = len(self.unexpectedSuccesses)

    def save_file_attachments(self, test, details):
        testcase_details = dict(
            content_attachments=[]
        )

        if details is not None:
            for name, detail in details.items():
                if detail.iter_bytes() != [b'']:
                    saved_name = create_unique_name(test.id(), name)
                    content = b''.join(chunk for chunk in detail.iter_bytes())
                    mime_type = str(detail.content_type)
                    self.data_store.put_file(saved_name, content, mime_type)
                    content_attachment = dict(
                        name=saved_name,
                        mime_type=mime_type,
                        size=len(content),
                        location=self.data_store.file_path(saved_name)
                    )
                    testcase_details['content_attachments'].append(
                        content_attachment
                    )
        self.testcase_details = testcase_details
        return testcase_details


def create_unique_name(test_id, file_name, timestamp=None):
    if timestamp is None:
        timestamp = datetime.timestamp(datetime.utcnow())
    slug = re.sub(r'(?u)[^-\w.]', '', file_name)
    return '%s+%s+%s' % (timestamp, test_id, slug)


def run_result(data_store, subunit_stream):
    suite = ByteStreamToStreamResult(subunit_stream)
    fs_result = FileSaveResult(data_store)
    result = StreamToExtendedDecorator(fs_result)
    result.startTestRun()
    try:
        suite.run(result)
    finally:
        result.stopTestRun()
    return True
