#!/usr/bin/env python3
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from datetime import datetime
import unittest
import os

from testtools import TestCase, TestResult
from testtools.content import Content
from testtools.content_type import UTF8_TEXT
from ucitests import fixtures as uci_fixtures

from subunitresults import subunit_parser
from subunitresults.ci_utils.testing import fixtures


def get_case(attachment_name, data):
    # Define the class in a funCtion so test loading won't try to load it as a
    # regular test class.
    class TestyTest(TestCase):
        def test_add_utf8_detail(self):
            self._timestamps = (
                datetime(2007, 12, 6, 15, 29, 43, 79060),
                datetime(2007, 12, 6, 15, 29, 43, 79060)
            )
            self.addDetail(
                attachment_name,
                Content(UTF8_TEXT, lambda: [data])
            )
    return TestyTest('test_add_utf8_detail')


class TestSubUnit(unittest.TestCase):

    def setUp(self):
        uci_fixtures.set_uniq_cwd(self)
        ds_path = os.path.join(self.uniq_dir, 'data_store')
        self.ds = fixtures.FakeDataStore(ds_path, None)
        self.addCleanup(self.ds.delete, recursive=True)

    def run_test_with_utf8_attachment(self, attachment_name, data):
        test = get_case(attachment_name, data)
        fs_result = subunit_parser.FileSaveResult(self.ds)
        test.run(fs_result)
        return fs_result

    def test_file_save_result(self):
        fs_result = subunit_parser.FileSaveResult(self.ds)
        self.assertIsInstance(fs_result, TestResult)

    def test_default_summary(self):
        fs_result = subunit_parser.FileSaveResult(self.ds)
        default_summary = dict(
            tests_run=0,
            errors=0,
            expected_failures=0,
            failures=0,
            num_attachments=0,
            run_time=0,
            skipped=0,
            unexpected_successes=0,
            tests=dict()
        )
        self.assertDictEqual(fs_result.summary, default_summary)

    def test_testsuite_summary(self):
        attachment_name = 'foo'
        attachment_data = b'bar'
        fs_result = self.run_test_with_utf8_attachment(
            attachment_name,
            attachment_data
        )
        expected_summary = {
            'failures': 0,
            'skipped': 0,
            'num_attachments': 1,
            'run_time': 0.0,
            'tests_run': 0,
            'errors': 0,
            'unexpected_successes': 0,
            'expected_failures': 0,
            'tests': {
                'tests.test_subunit_parser.TestyTest.test_add_utf8_detail':
                    ['pass', 0.0]
            }
        }
        self.assertDictEqual(fs_result.summary, expected_summary)

    def test_testcase_details(self):
        attachment_name = 'foo'
        attachment_data = b'bar'
        fs_result = self.run_test_with_utf8_attachment(
            attachment_name,
            attachment_data
        )
        self.assertEqual(fs_result.testcase_details['status'], 'pass')
        self.assertEqual(fs_result.testcase_details['run_time'], 0.0)
        attachment = fs_result.testcase_details['content_attachments'][0]
        self.assertEqual(attachment['size'], 3)

    def test_save_attachment(self):
        attachment_name = 'foo'
        attachment_data = b'bar'
        fs_result = self.run_test_with_utf8_attachment(
            attachment_name,
            attachment_data
        )
        attachments = fs_result.testcase_details['content_attachments']
        saved_name = attachments[0]['name']
        self.addCleanup(self.ds.delete_file, saved_name)
        self.assertEqual([saved_name], self.ds.list_files())
        self.assertEqual(attachment_data, self.ds.get_file(saved_name))

    def test_run_result(self):
        test_dir = os.path.dirname(os.path.realpath(__file__))
        subunit_file = os.path.join(test_dir, 'results_for_tests.subunit')
        with open(subunit_file, 'rb') as f:
            success = subunit_parser.run_result(self.ds, f)
        self.assertTrue(success)


class TestSubUnitName(unittest.TestCase):

    def test_create_unique_name(self):
        id = 'a.b.c'
        name = 'name'
        fake_epoch = '1407454141.539095'
        uniq_name = subunit_parser.create_unique_name(
            id,
            name,
            timestamp=fake_epoch
        )
        expected_name = '%s+%s+%s' % (fake_epoch, id, name)
        self.assertEqual(uniq_name, expected_name)
