#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
import subunit
from testtools import content
import unittest


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def run(*args, **kwargs):
    out = kwargs.pop('out', subprocess.PIPE)
    err = kwargs.pop('err', subprocess.PIPE)
    check_rc = kwargs.pop('check_rc', True)
    cwd = kwargs.pop('cwd', None)
    proc = subprocess.Popen(args, stdout=out, stderr=err, cwd=cwd)
    out, err = proc.communicate()
    if check_rc and proc.returncode:
        raise Exception(
            '{} failed with return code: {}'.format(' '.join(args),
                                                    proc.returncode))
    return proc, out, err


def parse_summary_line(line):
    try:
        name, status = line.split(None, 1)
    except ValueError:
        # Irregular line, set an arbitrary status not conflicting with the real
        # ones and report the whole line as wrong.
        return (line, 'error')
    if name == '*' and status.startswith('SKIP '):
        # Tweak a bit to normalize the status
        name = '*' + status[4:]
        status = 'SKIP'
    elif status.startswith('FAIL'):
        # Remove spurious info, more info is available in the details
        # (stdout/stderr for the corresponding test)
        status = 'FAIL'
    return (name, status)


def parse_summary(path):
    tests = []
    with open(path) as f:
        for line in f:
            line = line.strip()
            tests.append(parse_summary_line(line))
    return tests


def get_test_case(name):

    class Test(unittest.TestCase):

        def id(self):
            return name

    # We wont run that test, we only use it as a way to interact with subunit
    # in a simple and supported way.
    return Test('id')


def get_details(name):
    details = {}
    out_path = os.path.join('results', '{}-stdout'.format(name))
    if os.path.exists(out_path):
        with open(out_path) as f:
            details['stdout'] = content.text_content(f.read())
    err_path = os.path.join('results', '{}-stderr'.format(name))
    if os.path.exists(err_path):
        with open(err_path) as f:
            details['stderr'] = content.text_content(f.read())
    return details


def process_test_result(package, result, name, status):
    if status == 'SKIP':
        # adt-run doesn't give a name to the test here (there are none...) we
        # fake one.
        t = get_test_case('{}.skipped'.format(package))
    else:
        t = get_test_case(name)
    result.startTest(t)
    if status == 'PASS':
        result.addSuccess(t, details=get_details(name))
    elif status == 'FAIL':
        result.addFailure(t, details=get_details(name))
    elif status == 'SKIP':
        result.addSkip(t, reason=name)
    else:
        # We are catching unexpected lines in 'summary.log' here, some of them
        # may be valid (but yet unknown) cases. We'll fix them as they are
        # discovered. By turning them into an erroring test, we know the user
        # will see them and can report accordingly.
        error = 'adt-run produced an unexpected result: "{}"'.format(name)
        details = dict(error=content.text_content(error))
        result.addError(t, details=details)
    result.stopTest(t)


def run_test(ip, login, ssh_key_path, package):
    subunit_path = '{}.subunit'.format(package)
    output = open(subunit_path, 'wb')
    result = subunit.TestProtocolClient(output)
    result.startTestRun()
    try:
        summary_path = '{}-summary.log'.format(package)
        log_path = '{}.log'.format(package)
        cmd = ['adt-run',
               # Restrictions: build-needed will override --no-built-binaries
               # so it's safe to specify it.
               '--apt-source', package,
               '--summary', summary_path,
               '--log-file', log_path,
               '--output-dir', 'results',
               '---', 'adt-virt-ssh',
               '-H', ip, '-l', login, '-i', ssh_key_path]
        proc, out, err = run(*cmd, out=None, err=None, check_rc=False)
        rc = proc.returncode
        for name, status in parse_summary(summary_path):
            process_test_result(package, result, name, status)
    finally:
        result.stopTestRun()
        output.close()
    return rc
