#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import os


from ci_utils import (
    amqp_utils,
    amqp_worker,
)
from ci_utils import unit_config
from tstrun import (
    run_test,
    testbed,
)


class TestRunnerWorker(amqp_worker.AMQPWorker):

    def __init__(self, _data_store_factory=None):
        """Test runner worker.

        :param _data_store_factory: A callable returning a ticket specific data
            store (intended to be used by tests).
        """
        super(TestRunnerWorker, self).__init__('image_test')
        self.data_store_factory = _data_store_factory
        self.data_store = None
        # FIXME: This won't be needed anymore (at all) once
        # tests.test_test_runner.TestTestRunner.test_process_ticket is complete
        # and cover regressions. -- vila 2014-06-25
        if os.environ.get('HOME', None) is None:
            # Fall back to the current dir where the worker is run from
            os.environ['HOME'] = os.getcwd()

    def _create_data_store(self, ticket_id):
        if self.data_store_factory is None:
            self.data_store = super(TestRunnerWorker,
                                    self)._create_data_store(ticket_id)
        else:
            self.data_store = self.data_store_factory(ticket_id)
        return self.data_store

    def save_artifact(self, logger,  results, name, value, description=None,
                      kind='LOGS'):
        """Save an artifact catching and reporting exceptions.

        This should only be used to upload artifacts after the pass/fail status
        for the test run has been established so it's safe to fail the upload.

        :param logger: To report execution.

        :param results: The dict holding the 'artifacts' attribute.

        :param name: The artifact name.

        :param value: The artifact value as a string.

        :param description: For reporting purposes.

        :param kind: The kind of artifact (defaults to 'LOGS').
        """
        if description is None:
            description = name
        try:
            logger.info('Saving {}'.format(description))
            url = self.data_store.put_file(name, value, 'text/plain')
            self._create_artifact('{}.{}'.format(self.logger_name, name),
                                  url, results, kind)
        except:
            logger.exception('Unable to upload {}'.format(description))

    def save_testbed_console(self, logger, results, test_bed):
        self.save_artifact(logger, results,
                           'testbed-cloud-init.log',
                           test_bed.get_cloud_init_console(),
                           'testbed console')

    def save_testbed_artifacts(self, logger, results, test_bed, package):
        subunit_path = '{}.subunit'.format(package)
        self.save_artifact(
            logger, results,
            subunit_path, open(subunit_path).read(),
            'subunit results for {}'.format(package),
            'RESULTS')
        package_log_path = '{}.log'.format(package)
        self.save_artifact(
            logger, results,
            package_log_path, open(package_log_path).read(),
            'adt-run log for {}'.format(package))
        summary_log_path = '{}-summary.log'.format(package)
        self.save_artifact(
            logger, results,
            summary_log_path, open(summary_log_path).read(),
            'adt-run summary for {}'.format(package))
        syslog_log_path = '{}-syslog.log'.format(package)
        self.save_artifact(
            logger, results,
            syslog_log_path, test_bed.get_remote_content('/var/log/syslog'),
            'syslog for {}'.format(package))

    def test_package(self, test_bed, package):
        return run_test.run_test(test_bed.ip, 'ubuntu',
                                 test_bed.test_bed_key_path,
                                 package)

    def handle_request(self, params, logger):
        logger.info('Running in {} with HOME: {}'.format(os.getcwd(),
                                                         os.environ['HOME']))
        ticket_id = params['ticket_id']
        progress_queue = params['progress_trigger']
        series = params['series']
        image_id = params['image_id']
        ppa_list = params['ppa_list']
        package_list = params['package_list']

        results = {}

        def status_cb(msg):
            logger.info(msg)
            amqp_utils.progress_update(progress_queue, {'message': msg})

        self.data_store = self._create_data_store(params['ticket_id'])
        try:
            status_cb('Setting up the testbed for ticket {}'.format(ticket_id))
            auth_conf = unit_config.get_auth_config()
            tb_name = 'testbed-{}'.format(ticket_id)
            conf = testbed.vms_config_from_auth_config(tb_name, auth_conf)
            conf.set('vm.image', image_id)
            conf.set('vm.release', series)
            conf.set('vm.ppas', ','.join(ppa_list))
            base_dir = conf.get('vm.vms_dir')
            if not os.path.exists(base_dir):
                os.makedirs(base_dir)
            conf.store.save_changes()
            test_bed = testbed.TestBed(conf)
        except:
            logger.exception(
                'The testbed creation for {} failed'.format(ticket_id))
            return amqp_utils.progress_failed, results

        try:
            test_bed.setup()
            # Save the changes that can happen during the setup (mostly for
            # debug)
            conf.store.save_changes()
        except:
            logger.exception(
                'The testbed setup for {} failed'.format(ticket_id))
            if test_bed.instance is not None:
                self.save_testbed_console(logger, results, test_bed)
                test_bed.teardown()
            return amqp_utils.progress_failed, results

        self.save_testbed_console(logger, results, test_bed)
        status_cb('The test bed is ready')
        # The tests will succeed unless they fail ;)
        notify = amqp_utils.progress_completed
        try:
            for package in package_list:
                if params.get('cancelled'):
                    logger.error('The request for {} has been cancelled,'
                                 ' exiting'.format(ticket_id))
                    return amqp_utils.progress_failed, results
                status_cb('Testing {}'.format(package))
                # uci-vms shell adt-run ...  --- adt-virt-null
                return_code = self.test_package(test_bed, package)
                # 0 is success, 2 means "at least one test skipped" and is
                # considered a success
                if return_code not in (0, 2):
                    # At least one test failed
                    notify = amqp_utils.progress_failed
                self.save_testbed_artifacts(logger, results, test_bed, package)
            status_cb('Test completed for ticket {}'.format(ticket_id))
            return notify, results
        except:
            logger.exception(
                'Exception while handling ticket {}'.format(ticket_id))
            raise
        finally:
            test_bed.teardown()


if __name__ == '__main__':
    TestRunnerWorker().main(amqp_utils.TEST_RUNNER_QUEUE)
