#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import logging
import os
import requests
import subprocess
import time


from novaclient import exceptions
from novaclient.v1_1 import client
from uciconfig import options
from ucivms import (
    config,
    vms,
)

from ci_utils import unit_config


def register(option):
    options.option_registry.register(option)


register(options.Option('os.username', default_from_env='GLANCE_OS_USERNAME',
                        default=options.MANDATORY,
                        help_string='''The Open Stack user name.

This is generally set via GLANCE_OS_USERNAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.password', default_from_env='GLANCE_OS_PASSWORD',
                        default=options.MANDATORY,
                        help_string='''The Open Stack password.

This is generally set via GLANCE_OS_PASSWORD, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.region_name',
                        default_from_env='GLANCE_OS_REGION_NAME',
                        default=options.MANDATORY,
                        help_string='''The Open Stack region name.

This is generally set via GLANCE_OS_REGION_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.tenant_name',
                        default_from_env='GLANCE_OS_TENANT_NAME',
                        default=options.MANDATORY,
                        help_string='''The Open Stack tenant name.

This is generally set via GLANCE_OS_TENANT_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.auth_url', default_from_env='GLANCE_OS_AUTH_URL',
                        default=options.MANDATORY,
                        help_string='''The Open Stack keystone url.

This is generally set via GLANCE_OS_AUTH_URL, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.ListOption('vm.flavors', default=None,
                            help_string='''\
A list of flavors for all supported clouds.

The first known one is used.
'''))
register(options.Option('vm.image', default=options.MANDATORY,
                        help_string='''The base image to run the tests.'''))
register(config.PathOption('vm.ssh_key_path', default='~/.ssh/id_rsa',
                           help_string='''The ssh key for the testbed.'''))
register(options.ListOption('vm.apt_get.update.timeouts',
                            default='15.0, 90.0, 240.0',
                            help_string='''apt-get update timeouts in seconds.

When apt-get update fails on hash sum mismatches, retry after the specified
timeouts. More values mean more retries.
'''))
register(options.ListOption('vm.ppas',
                            default=None,
                            help_string='''PPAs to be added to the testbed.

This works around cloud-init not activating the deb-src line and not providing
a way to do so. This is intended to be fixed in uci-vms so vm.apt_sources can
be used again.
'''))


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


HERE = os.path.abspath(os.path.dirname(__file__))


class TestBedException(Exception):
    pass


def vms_config_from_auth_config(testbed_name, auth_conf):
    """Populate a ucivms config from an auth config.

    :param testbed_name: The name of the test bed in the ucivms config file.

    :param auth_conf: A dict with the openstack credentials.
    """
    conf = config.VmStack(testbed_name)
    conf.set('vm.name', testbed_name)
    conf.set('os.username', auth_conf.get('glance_auth_user'))
    conf.set('os.password', auth_conf.get('glance_auth_password'))
    conf.set('os.tenant_name', auth_conf.get('glance_auth_tenant_name'))
    conf.set('os.region_name', auth_conf.get('glance_auth_region'))
    conf.set('os.auth_url', auth_conf.get('glance_auth_url'))
    conf.set('vm.flavors', ','.join(auth_conf.get('tr_flavors')))
    # User can override locally in unit_config
    conf.set('vm.ssh_key_path',
             unicode(auth_conf.get('auth_testbed_sshkey',
                                   '~/.ssh/id_rsa')))
    return conf


class TestBed(vms.VM):

    def __init__(self, conf):
        super(TestBed, self).__init__(conf)
        self.instance = None
        self.floating_ip = None
        self.nova = self.build_nova_client()
        self.test_bed_key_path = None
        self.conf.set('vm.ssh_authorized_keys',
                      self.conf.get('vm.ssh_key_path') + '.pub')
        # No need to reboot a nova instance
        self.conf.set('vm.poweroff', 'False')
        self.conf.set('vm.final_message', 'testbed setup completed.')

    def build_nova_client(self):
        args = [self.conf.get('os.username'),
                self.conf.get('os.password'),
                self.conf.get('os.tenant_name'),
                self.conf.get('os.auth_url')]
        kwargs = {'region_name': self.conf.get('os.region_name'),
                  'service_type': 'compute'}
        return client.Client(*args, **kwargs)

    def ensure_ssh_key_is_available(self):
        self.test_bed_key_path = self.conf.get('vm.ssh_key_path')
        # From the test runner, we need an ssh key that can be used to connect
        # to the test bed. For testing purposes, we rely on self.auth_conf to
        # provide this key.
        if not os.path.exists(self.test_bed_key_path):
            base_dir = os.path.dirname(self.test_bed_key_path)
            if not os.path.exists(base_dir):
                os.makedirs(base_dir)
            # First time the test runner instance needs to create a test bed
            # instance, we need to create the ssh key pair.
            subprocess.call(
                ['ssh-keygen', '-t', 'rsa', '-q',
                 '-f', self.test_bed_key_path, '-N', ''])

    def find_flavor(self):
        for flavor in self.conf.get('vm.flavors'):
            try:
                return self.nova.flavors.find(name=flavor)
            except exceptions.NotFound:
                pass
        # MISSINGTEST: some cloud doesn't provide one of our expected flavors
        # -- vila 2014-01-06
        raise TestBedException(
            'None of {} can be found'.format(self.flavors))

    def find_nova_image(self):
        image_name = self.conf.get('vm.image')
        try:
            return self.nova.images.find(name=image_name)
        except exceptions.NotFound:
            raise TestBedException(
                'Image "{}" cannot be found'.format(image_name))

    def setup(self):
        flavor = self.find_flavor()
        image = self.find_nova_image()
        self.ensure_ssh_key_is_available()
        self.create_user_data()
        with open(self._user_data_path) as f:
            user_data = f.read()
        self.instance = self.nova.servers.create(
            name=self.conf.get('vm.name'), flavor=flavor, image=image,
            userdata=user_data)
        self.wait_for_active_instance()
        if unit_config.is_hpcloud(self.conf.get('os.auth_url')):
            self.floating_ip = self.nova.floating_ips.create()
            self.instance.add_floating_ip(self.floating_ip)
        self.wait_for_ip()
        self.wait_for_cloud_init()
        ppas = self.conf.get('vm.ppas')
        if ppas:
            cmd = ['sudo', 'add-apt-repository']
            if self.conf.get('vm.release') > 'precise':
                cmd.append('--enable-source')
            for ppa in ppas:
                self.ssh(*(cmd + [ppa]))
        # Now we can apt-get update (doing it earlier would lead to the wrong
        # source package to be installed).
        self.safe_apt_get_update()

    def apt_get_update(self):
        return self.ssh('sudo', 'apt-get', 'update')

    def safe_apt_get_update(self):
        for timeout in self.conf.get('vm.apt_get.update.timeouts'):
            proc, out, err = self.apt_get_update()
            if proc.returncode == 0:
                # We're done
                return
            else:
                msg = ('apt-get update failed, wait {}s'
                       'stdout:\n{}\n'
                       'stderr:\n{}\n')
                log.info(msg.format(timeout, out, err))
                time.sleep(float(timeout))
        raise TestBedException('apt-get update never succeeded')

    def update_instance(self):
        # MISSINGTEST: What if the instance disappear ? (could be approximated
        # by deleting the nova instance) -- vila 2014-06-05
        try:
            # Always query nova to get updated data about the instance
            self.instance = self.nova.servers.get(self.instance.id)
            return True
        except requests.ConnectionError:
            # The reported status is the one known before the last attempt.
            log.warn('Received connection error for {},'
                     ' retrying (status was: {})'.format(
                         self.instance.id, self.instance.status))
        return False

    def wait_for_active_instance(self):
        # FIXME: get_active_instance should be a config option
        # -- vila 2014-05-13
        get_active_instance = 300  # in seconds so 5 minutes
        timeout_limit = time.time() + get_active_instance
        while time.time() < timeout_limit and self.instance.status != 'ACTIVE':
            time.sleep(5)
            self.update_instance()
        if self.instance.status != 'ACTIVE':
            # MISSINGTEST: What if the instance doesn't come up ?
            msg = 'Instance never came up (last status: {})'.format(
                self.instance.status)
            raise TestBedException(msg)

    def wait_for_ip(self):
        # FIXME: get_ip_timeout should be a config option -- vila 2014-01-30
        get_ip_timeout = 300  # in seconds so 5 minutes
        timeout_limit = time.time() + get_ip_timeout
        while time.time() < timeout_limit:
            if not self.update_instance():
                time.sleep(5)
                continue
            networks = self.instance.networks.values()
            if networks:
                # The network name is arbitrary, can vary for different clouds
                # but there should be only one network so we get the first one
                # and avoid the need for a config option for the network name.
                # We take the last IP address so it's either the only one or
                # the floating one. In both cases that gives us a reachable IP.
                self.ip = networks[0][-1]
                log.info('Got IP {} for {}'.format(self.ip, self.instance.id))
                # FIXME: Right place to report how long it took to spin up the
                # instance as far as nova is concerned. -- vila 2014-01-30
                return
            else:
                log.info(
                    'IP not yet available for {}'.format(self.instance.id))
            time.sleep(5)
        # MISSINGTEST: What if the instance still doesn't have an ip ?
        msg = 'Instance {} never provided an IP'.format(self.instance.id)
        raise TestBedException(msg)

    def get_cloud_init_console(self, length=None):
        return self.instance.get_console_output(length)

    def wait_for_cloud_init(self):
        # FIXME: cloud_init_timeout should be a config option (related to
        # get_ip_timeout and probably the two can be merged) -- vila 2014-01-30
        cloud_init_timeout = 1200  # in seconds so 20 minutes
        timeout_limit = time.time() + cloud_init_timeout
        final_message = self.conf.get('vm.final_message')
        while time.time() < timeout_limit:
            # A relatively cheap way to catch cloud-init completion is to watch
            # the console for the specific message we specified in user-data).
            try:
                console = self.get_cloud_init_console(10)
            except exceptions.OverLimit:
                # This happens rarely but breaks badly if not caught. elmo
                # recommended a 30 seconds nap in that case.
                nap_time = 30
                msg = ('Rate limit while acquiring nova console for {},'
                       ' will sleep for {} seconds')
                log.exception(msg.format(self.instance.id, nap_time))
                time.sleep(nap_time)
                continue
            except requests.ConnectionError:
                # The reported status is the one known before the last attempt.
                log.warn('Received connection error for {},'
                         ' retrying)'.format(self.instance.id))
                time.sleep(5)
                continue
            if final_message in console:
                # We're good to go
                log.info(
                    'cloud-init completed for {}'.format(self.instance.id))
                return
            time.sleep(5)
        raise TestBedException('Instance never completed cloud-init')

    def teardown(self):
        if self.instance is not None:
            self.nova.servers.delete(self.instance.id)
        if self.floating_ip is not None:
            self.nova.floating_ips.delete(self.floating_ip)
        # FIXME: Now we can remove the testbed key from known_hosts (see
        # ssh()). -- vila 2014-01-30

    def ssh(self, command, *args, **kwargs):
        out = kwargs.pop('out', None)
        err = kwargs.pop('err', None)
        user = 'ubuntu'
        host = self.ip
        cmd = ['ssh',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               '{}@{}'.format(user, host)]
        if command is not None:
            cmd += [command]
        if args:
            cmd += args
        proc = subprocess.Popen(cmd, stdout=out, stderr=err)
        out, err = proc.communicate()
        return proc, out, err

    def scp(self, local_path, remote_path):
        cmd = ['scp',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               local_path, remote_path]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        return proc, out, err

    def get_remote_content(self, path):
        proc, content, err = self.ssh('cat', path, out=subprocess.PIPE,
                                      err=subprocess.PIPE)
        if proc.returncode:
            # We didn't get a proper content, report it instead
            content = ("{} couldn't be copied from testbed {}:\n"
                       "error: {}\n".format(path, self.ip, err))
        return content

    def run_test(self, package):
        proc, _, _ = self.ssh('bin/run_test.py', package)
        return proc.returncode
