# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cStringIO import StringIO
import logging
import os
import unittest

from ucitests import (
    assertions,
)
from ucivms.tests import fixtures as vms_fixtures
from ci_utils import (
    amqp_utils,
    unit_config,
)
from ci_utils.testing import (
    features,
    fixtures,
)
from tstrun import (
    run_worker,
    tests,
)


@features.requires(tests.nova_creds)
@features.requires(features.nova_compute)
class TestWorkerHandleRequest(unittest.TestCase):

    def setUp(self):
        super(TestWorkerHandleRequest, self).setUp()
        # A logger for the test runner worker
        self.logger_stream = logging.StreamHandler(StringIO())
        self.logger_stream.setFormatter(logging.Formatter('%(message)s'))
        self.logger = logging.getLogger()
        self.logger.addHandler(self.logger_stream)
        # The fake rabbit input queue
        self.queue = fixtures.FakeTaskQueue('test')
        # The data store and uci-vms need disk access
        vms_fixtures.isolate_from_disk(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        # We don't need a config for the data store
        self.data_store = fixtures.FakeDataStore(self.ds_path)

        def create_data_store(component):
            return self.data_store

        self.ds_factory = create_data_store

    def get_ppa_list(self):
        # We're testing only the test runner so there is no ppa assigned to the
        # ticket, only the master ppa is relevant.
        conf = unit_config.get_auth_config()
        return [conf.get('master_ppa')]

    def assertTestRun(self, params):
        worker = run_worker.TestRunnerWorker(self.ds_factory)
        self.queue.publish(params)
        returns = self.queue.subscribe(worker.handle_request, self.logger)
        assertions.assertLength(self, 1, returns)
        self.assertTrue(self.queue.is_empty())
        self.assertEquals(['libpng-summary.log', 'libpng-syslog.log',
                           'libpng.log', 'libpng.subunit',
                           'testbed-cloud-init.log'],
                          self.data_store.list_files())
        return returns

    def test_handle_request_precise_image(self):
        # FIXME: We need a way to get an image that is available to the nova
        # setup we use. -- vila 2014-05-12
        params = dict(ticket_id='precise', progress_trigger='progress',
                      series='precise',
                      image_id=('Ubuntu 12.04.4 LTS (amd64 20140613)'
                                ' - CI Engineering'),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng'])
        returns = self.assertTestRun(params)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_failed, retcode)
        # We need to check the content here as the worker will put an error
        # message there if it cannot get them from the testbed (so the user get
        # *some* feedback).
        self.assertEquals(
            '*                    SKIP no tests in this package\n',
            self.data_store.get_file('libpng-summary.log'))

    def test_handle_request_saucy_image(self):
        # FIXME: We need a way to get an image that is available to the nova
        # setup we use. -- vila 2014-05-12
        params = dict(ticket_id='saucy', progress_trigger='progress',
                      series='saucy',
                      image_id=('Ubuntu Server 13.10 (amd64 20140409.1)'
                                ' - Partner Image'),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng'])
        returns = self.assertTestRun(params)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_completed, retcode)
        # We need to check the content here as the worker will put an error
        # message there if it cannot get them from the testbed (so the user get
        # *some* feedback).
        self.assertEquals('build                PASS\n',
                          self.data_store.get_file('libpng-summary.log'))

    def test_handle_request_trusty_image(self):
        # FIXME: We need a way to get an image that is available to the nova
        # setup we use. -- vila 2014-05-12
        params = dict(ticket_id='trusty', progress_trigger='progress',
                      series='trusty',
                      image_id=('Ubuntu Server 14.04 LTS (amd64 20140607.1)'
                                ' - Partner Image'),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng'])
        returns = self.assertTestRun(params)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_completed, retcode)
        # We need to check the content here as the worker will put an error
        # message there if it cannot get them from the testbed (so the user get
        # *some* feedback).
        self.assertEquals('build                PASS\n',
                          self.data_store.get_file('libpng-summary.log'))
