# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import collections
import errno
import os
import subprocess
import sys
import logging

logging.basicConfig(level=logging.INFO, format='%(msg)s')
log = logging.getLogger('integration-test-runner-deployers')

HERE = os.path.abspath(os.path.dirname(__file__))
# We need access to ci_utils
sys.path.append(os.path.join(HERE, '..', 'ci-utils'))
# We need to find our juju environment in deploy
sys.path.append(os.path.join(HERE, '..', 'juju-deployer'))
import deploy

from ci_utils import testing
from ci_utils.testing import features


def _juju_call(juju_args):
    if not isinstance(juju_args, collections.Iterable):
        raise OSError(
            'Given argument is not iterable: {!r}'.format(juju_args))
    args = ['juju']
    args.extend(juju_args)
    try:
        p = subprocess.Popen(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        # MISSINGTEST: not covered by any test -- vila 2014-06-20
        if e.errno != errno.ENOENT:
            raise
        raise OSError("juju not found, do you have Juju installed?")
    out, err = p.communicate()
    if p.returncode:
        raise IOError('juju command failed {!r}:\n{}'.format(juju_args, err))
    return out


def juju_run(service, cmd, unit=0):
    """Call a command on a single given unit.

    :param service: A service name.

    :param cmd: The command to execute on a unit. This should be provided as a
        string to avoid evaluation on the originating host (the evaluation will
        happen on the unit itself if needed).

    :param unit: The unit number where the command should be run. It defaults
        to 0 if not provided.

    """
    return _juju_call(['run', '--unit', '{}/{}'.format(service, unit), cmd])


def juju_expose(service, expose=True):
    if expose:
        action = 'expose'
    else:
        action = 'unexpose'
    return _juju_call([action, service])


# TODO ev 2014-06-16 @features.requires(features.settled_units)
@features.requires(features.bootstrapped_juju)
class DeployerTest(testing.TestCaseWithGnupg):
    '''Base class for building juju deployer based tests.'''

    def get_ip_and_port(self, service, unit=0):
        status = deploy.juju_status()
        units = status['services'][service]['units']
        ip = units['%s/%d' % (service, unit)]['public-address']
        formatted = '%s/%d' % (service, unit)
        port, _ = units[formatted]['open-ports'][0].split('/')
        return ip, int(port)

    def assert_job_running(self, service, unit=0, upstart_job=None):
        # Ensure the given upstart job is running on a unit
        if upstart_job is None:
            upstart_job = service
        output = juju_run(service, '/sbin/status {}'.format(upstart_job),
                          unit=unit)
        self.assertIn('start/running', output)
