# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import httplib2
import json
import logging
import socket
import uuid

from django.shortcuts import get_object_or_404

from tastypie import fields
from tastypie.constants import ALL, ALL_WITH_RELATIONS
from tastypie.authorization import Authorization
from tastypie.resources import ModelResource, Resource, Bundle
from ticket.models import (Review, Ticket, SubTicket, SourcePackageUpload,
                           TicketArtifact, SubTicketArtifact,
                           SubTicketWorkflowStep, SubTicketWorkflowStepStatus,
                           TicketWorkflowStep, TicketWorkflowStepStatus,
                           get_enum_title, get_enum_value)

from project.models import BinaryPackage
from project.api import SourcePackageResource

from ci_utils import ticket_states
from ci_utils.json_status import JSONStatus


class TicketTranslatedResource(ModelResource):

    def dehydrate_current_workflow_step(self, bundle):
        return get_enum_title(bundle.data["current_workflow_step"],
                              TicketWorkflowStep)

    def hydrate_current_workflow_step(self, bundle):
        value = bundle.data.get('current_workflow_step')
        if not value:
            return bundle
        try:
            int(value)
        except ValueError:
            # user passed us "Package building"
            value = get_enum_value(value, TicketWorkflowStep)
            bundle.data['current_workflow_step'] = value
        return bundle

    def dehydrate_status(self, bundle):
        return get_enum_title(bundle.data["status"], TicketWorkflowStepStatus)

    def hydrate_status(self, bundle):
        value = bundle.data.get('status')
        if not value:
            return bundle
        try:
            int(value)
        except ValueError:
            # user passed us "Continue"
            value = get_enum_value(value, TicketWorkflowStepStatus)
            bundle.data['status'] = value
        return bundle


class SubTicketTranslatedResource(ModelResource):

    def dehydrate_current_workflow_step(self, bundle):
        return get_enum_title(bundle.data["current_workflow_step"],
                              SubTicketWorkflowStep)

    def dehydrate_status(self, bundle):
        return get_enum_title(bundle.data["status"],
                              SubTicketWorkflowStepStatus)


class TicketResource(TicketTranslatedResource):

    class Meta:
        queryset = Ticket.objects.filter(private=False)
        allowed_methods = ['get', 'post']
        authorization = Authorization()

    def dispatch(self, request_type, request, **kwargs):
        """Support Ticket lookup/traversal by ID or UUID."""
        pk = kwargs.get('pk')
        try:
            uuid.UUID(pk)
            try:
                kwargs['pk'] = Ticket.objects.get(uuid=pk).pk
            except Ticket.DoesNotExist:
                # Delegate the 404 to tastypie.
                kwargs['pk'] = 0
        except (TypeError, ValueError):
            # It's not a UUID, proceed with tastypie lookup.
            pass

        return super(TicketResource, self).dispatch(
            request_type, request, **kwargs)


class OpenTicketResource(TicketTranslatedResource):

    class Meta:
        queryset = Ticket.objects.all()
        allowed_methods = ['get']
        resource_name = 'opentickets'

    def get_object_list(self, request):
        exclude = [TicketWorkflowStep.NEW.value,
                   TicketWorkflowStep.COMPLETED.value]
        return super(OpenTicketResource, self).get_object_list(
            request).exclude(current_workflow_step__in=exclude).order_by('id')


class SourcePackageUploadResource(ModelResource):
    sourcepackage = fields.ToOneField(SourcePackageResource, 'sourcepackage',
                                      full=True)

    class Meta:
        queryset = SourcePackageUpload.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()
        resource_name = 'spu'


class SubTicketResource(SubTicketTranslatedResource):
    ticket = fields.ToOneField(TicketResource, 'ticket', full=True)
    source_package_upload = fields.ToOneField(SourcePackageUploadResource,
                                              'source_package_upload',
                                              full=True)

    class Meta:
        queryset = SubTicket.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()


class TicketArtifactResource(ModelResource):
    ticket = fields.ToOneField(TicketResource, 'ticket', null=True, full=True)

    class Meta:
        queryset = TicketArtifact.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()
        filtering = {
            'type': ['exact'],
            'ticket': ['exact'],
        }


class SubTicketArtifactResource(ModelResource):
    subticket = fields.ToOneField(SubTicketResource, 'subticket', null=True,
                                  full=True)

    class Meta:
        queryset = SubTicketArtifact.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()


class FullTicketArtifactResource(ModelResource):

    class Meta:
        queryset = TicketArtifact.objects.all()
        allowed_methods = ['get']


class FullSubTicketArtifactResource(ModelResource):

    class Meta:
        queryset = SubTicketArtifact.objects.all()
        allowed_methods = ['get']


class FullSubTicketResource(SubTicketTranslatedResource):
    source_package_upload = fields.ToOneField(SourcePackageUploadResource,
                                              'source_package_upload',
                                              full=True)
    artifact = fields.ToManyField(FullSubTicketArtifactResource,
                                  'subticketartifact_set', full=True)

    class Meta:
        queryset = SubTicket.objects.all()
        allowed_methods = ['get']


class FullTicketResource(TicketTranslatedResource):
    subticket = fields.ToManyField(FullSubTicketResource, 'subticket_set',
                                   full=True)
    artifact = fields.ToManyField(FullTicketArtifactResource,
                                  'ticketartifact_set', null=True,
                                  full=True)

    class Meta:
        queryset = Ticket.objects.all()
        allowed_methods = ['get']


class TicketUpdateStatusResource(TicketTranslatedResource):

    class Meta:
        queryset = Ticket.objects.all()
        fields = ['id', 'current_workflow_step', 'status', 'lander_unit']
        allowed_methods = ['get', 'patch']
        authorization = Authorization()
        resource_name = 'updateticketstatus'
        filtering = {
            "id": ('exact'),
        }

    def patch_detail(self, request, **kwargs):
        update = json.loads(request.body)
        step = update.get("current_workflow_step")
        status = update.get("status")
        if (
            step and status and
            step == TicketWorkflowStep.COMPLETED.value and
            status == TicketWorkflowStepStatus.COMPLETED.value
        ):
            queryset = get_object_or_404(Ticket, pk=kwargs['pk'])
            added_binaries = queryset.added_binaries
            if added_binaries is not None:
                added_binaries = added_binaries.split(",")
            else:
                added_binaries = []
            removed_binaries = queryset.removed_binaries
            if removed_binaries is not None:
                removed_binaries = removed_binaries.split(",")
            else:
                removed_binaries = []
            for binary in added_binaries:
                #Use get or create incase the binary already exists in the list
                BinaryPackage.objects.get_or_create(name=binary)
            for binary in removed_binaries:
                try:
                    # If the binary specified exists in the list, delete it
                    bp = BinaryPackage.objects.get(name=binary)
                    bp.delete()
                except BinaryPackage.DoesNotExist:
                    # If the binary specified does not exist, move on
                    pass
        return super(TicketUpdateStatusResource, self).patch_detail(request,
                                                                    **kwargs)


class TicketStatusResource(TicketTranslatedResource):

    class Meta:
        queryset = Ticket.objects.all()
        fields = ['id', 'title', 'current_workflow_step', 'status', 'created',
                  'updated']
        allowed_methods = ['get']
        filtering = {
            "current_workflow_step": ALL,
            "status": ALL,
        }


class SubTicketUpdateStatusResource(SubTicketTranslatedResource):

    class Meta:
        queryset = SubTicket.objects.all()
        fields = ['id', 'current_workflow_step', 'status']
        allowed_methods = ['get', 'patch']
        authorization = Authorization()
        resource_name = 'updatesubticketstatus'
        filtering = {
            "id": ('exact'),
        }


class SubTicketStatusResource(SubTicketTranslatedResource):

    class Meta:
        queryset = SubTicket.objects.all()
        fields = ['id', 'title', 'current_workflow_step', 'status']
        allowed_methods = ['get']
        filtering = {
            "current_workflow_step": ALL,
        }


class WorkflowItem(object):
    def __init__(self, initial=None):
        self.__dict__['_data'] = {}

        if hasattr(initial, 'items'):
            self.__dict__['_data'] = initial
        else:
            self.__dict__['_data']['name'] = 'unknown'
            self.__dict__['_data']['value'] = 0

    def __getattr__(self, name):
        return self._data.get(name, None)

    def __setattr__(self, name, value):
        self.__dict__['_data'][name] = value

    def to_dict(self):
        return self._data


class WorkflowResource(Resource):
    name = fields.CharField(attribute='name')
    key = fields.CharField(attribute='key')
    value = fields.IntegerField(attribute='value')

    enum_type = None

    class Meta:
        abstract = True
        allowed_methods = ['get']
        object_class = WorkflowItem

    def get_object_list(self, request):
        results = []
        for item in self.enum_type:
            name = item.title
            key = item.token
            try:
                value = getattr(self.enum_type, key).value
            except AttributeError:
                continue

            wfs = WorkflowItem(
                initial={
                    'name': name,
                    'value': value,
                    'key': key,
                }
            )
            results.append(wfs)

        return results

    def obj_get_list(self, bundle, **kwargs):
        return self.get_object_list(bundle.request)

    def obj_get(self, bundle, **kwargs):
        key = kwargs['pk']

        result = {}
        try:
            value = getattr(self.enum_type, key).value
            name = getattr(self.enum_type, key).title
            result = WorkflowItem(
                initial={
                    'name': name,
                    'value': value,
                    'key': key,
                }
            )
        except AttributeError:
            pass

        return result

    def detail_uri_kwargs(self, bundle_or_obj):
        kwargs = {}

        if isinstance(bundle_or_obj, Bundle):
            kwargs['pk'] = bundle_or_obj.obj.key
        else:
            kwargs['pk'] = bundle_or_obj.key

        return kwargs


class TicketWorkflowStepResource(WorkflowResource):

    enum_type = ticket_states.TicketWorkflowStep

    class Meta(WorkflowResource.Meta):
        resource_name = 'ticket_workflow_step'


class SubTicketWorkflowStepResource(WorkflowResource):

    enum_type = ticket_states.SubTicketWorkflowStep

    class Meta(WorkflowResource.Meta):
        resource_name = 'subticket_workflow_step'


class TicketWorkflowStepStatusResource(WorkflowResource):

    enum_type = ticket_states.TicketWorkflowStepStatus

    class Meta(WorkflowResource.Meta):
        resource_name = 'ticket_workflow_step_status'


class SubTicketWorkflowStepStatusResource(WorkflowResource):

    enum_type = ticket_states.SubTicketWorkflowStepStatus

    class Meta(WorkflowResource.Meta):
        resource_name = 'subticket_workflow_step_status'


class ReviewResource(ModelResource):
    ticket = fields.ForeignKey(TicketResource, 'ticket')

    class Meta:
        queryset = Review.objects.all()
        authorization = Authorization()
        filtering = {
            'ticket': ALL_WITH_RELATIONS,
            'review_type': ALL,
            'completed': ALL,
        }


class TSStatus(ModelResource):
    class Meta():
        resource_name = 'status'
        allowed_methods = ['get']
        # return an empty queryset to make tastypie happy

    def get_list(self, request, **kwargs):
        status = JSONStatus()
        status.add_rabbit_configured()
        status.add_deployed_revision()
        create_check = self._check_create_ticket()
        status.add_true_false('Ticket Creation Works', create_check,
                              create_check is True)
        return self.create_response(request, status.results)

    def _get_ip(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("launchpad.net", 80))
            ip = s.getsockname()[0]
            s.close()
        except:
            logging.exception('ERROR: unable to determine local IP')
            ip = 'IP lookup error'
        return ip

    def _check_create_ticket(self):
        ip = self._get_ip()
        if ip == 'IP lookup error':
            return ip
        url = 'http://{}:8080/api/v1/ticket/'.format(ip)
        ticket_content = {"owner": "create@example.com",
                          "title": "Created ticket",
                          "description": "This ticket is for creation test",
                          "bug_id": "111111",
                          "added_binaries": "binary-1,binary-2",
                          "removed_binaries": "binary-3,binary-4"}
        client = httplib2.Http()
        resp, content = client.request(
            uri=url,
            method='POST',
            headers={'Content-Type': 'application/json'},
            body=json.dumps(ticket_content),
        )
        ticket_id = resp['location'].split('/')[-2]
        Ticket.objects.get(id=ticket_id).delete()
        return resp['status'] == '201'
