# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import uuid

from django.core.exceptions import ValidationError
from django.db import models, IntegrityError
from django.conf import settings
from project.models import SourcePackage

from ci_utils import amqp_utils, SUPPORTED_SERIES
from ci_utils.data_store import create_for_ticket
from ci_utils.unit_config import get_auth_config
from ci_utils.ticket_states import (
    STEP_TO_WORKFLOW_MAPPING,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


def _choices(enum):
    return [
        (int(item.value), item.title) for item in enum.items]


def get_enum_title(value, enumcls):
    return enumcls.items.mapping[value].title


def get_enum_value(title, enumcls):
    for item in enumcls.items:
        if item.title == title:
            return item.value


def get_ticket_uuid():
    """UUIDv1 factory for tickets.

    Returns the textual representation of a new UUIDv1.
    """
    return str(uuid.uuid1())


class Workflow(models.Model):
    '''A ticket is configured with a workflow. This specifies what steps the
    lander needs to handle. We store these in a table so that we can
    add/remove/update workflows in production without having to do an upgrade.
    '''
    name = models.CharField(
        primary_key=True,
        max_length=4096,
        help_text='A human readable label')
    steps = models.CharField(
        max_length=4096,
        help_text='JSON encoded data that specifying the lander steps')

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        '''ensure we don't allow invalid json'''
        try:
            json.loads(self.steps)
        except:
            raise ValidationError('Invalid JSON provided for "steps"')
        return super(Workflow, self).save(*args, **kwargs)

    @staticmethod
    def initialize():
        steps = [
            {'name': 'ppa_assigner'},
            {'name': 'bsbuilder'},
            {'name': 'image_builder'},
            {'name': 'test_runner'},
            {'name': 'publisher'},
            {'name': 'ppa_assigner_release'},
        ]
        if Workflow.objects.filter(name='default').count() == 0:
            Workflow.objects.create(name='default', steps=json.dumps(steps))


class Ticket(models.Model):
    class Meta:
        db_table = 'ticket'

    SERIES_CHOICES = tuple(
        (series, series.upper()) for series in SUPPORTED_SERIES)

    uuid = models.CharField(
        max_length=36, editable=False, unique=True, default=get_ticket_uuid,
        help_text=('UUIDv1 ticket identification represented as '
                   'string (aka \'hyphenated\').'))
    owner = models.EmailField(max_length=254)
    title = models.CharField(max_length=4096)
    description = models.TextField()
    bug_id = models.IntegerField(null=True, blank=True)
    current_workflow_step = models.IntegerField(
        choices=_choices(TicketWorkflowStep), max_length=4096,
        default=000)
    status = models.IntegerField(
        choices=_choices(TicketWorkflowStepStatus), max_length=4096,
        default=000)
    base_image = models.CharField(max_length=4096,
                                  default=settings.BASE_IMAGE_DEFAULT)
    series = models.CharField(max_length=4096,
                              choices=SERIES_CHOICES,
                              default=settings.SERIES_DEFAULT)
    master_ppa = models.CharField(max_length=4096,
                                  default=settings.MASTER_PPA_DEFAULT)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    # Comma-separated list of binary names to install or not install when
    # building the image for ticket testing.
    added_binaries = models.CharField(max_length=4096, null=True,
                                      blank=True)
    removed_binaries = models.CharField(max_length=4096, null=True,
                                        blank=True)
    workflow = models.ForeignKey(Workflow, default='default')

    lander_unit = models.IntegerField(
        default=0,
        help_text='The lander\'s juju unit number processing this ticket')

    private = models.BooleanField(default=False)

    def __unicode__(self):
        return self.title

    def create_container(self):
        ds = create_for_ticket(self.id, get_auth_config())
        ds.ensure_swift_client()

    def save(self, *args, **kwargs):
        """Overridden to validate added/removed package list formatting."""

        def is_valid_package_list(package_list_str):
            if package_list_str is None:
                return True
            if " " in package_list_str:
                raise IntegrityError("Added and removed packages have to be "
                                     "comma-separated lists with no spaces.")
            return len(package_list_str.split(",")) > 0

        if (not is_valid_package_list(self.added_binaries) or
                not is_valid_package_list(self.removed_binaries)):
            raise IntegrityError("Added and removed packages have to be "
                                 "comma-separated lists with no spaces.")

        is_new = False
        if int(self.current_workflow_step) == int(TicketWorkflowStep.NEW):
            # Block *updates* to NEW.
            if self.pk is not None:
                raise IntegrityError('Cannot update a ticket back to NEW.')
            is_new = True

        super(Ticket, self).save(*args, **kwargs)

        # Create Swift ticket container when the ticket is created.
        # It has to be done after saving because it need the ticket ID.
        if is_new:
            self.create_container()
        self._handle_lander()

    def _handle_lander(self):
        notify = False
        if int(self.current_workflow_step) == int(TicketWorkflowStep.QUEUED):
            notify = (int(self.status) == int(TicketWorkflowStepStatus.QUEUED))

        resume = (int(self.status) == int(TicketWorkflowStepStatus.CONTINUE))

        if notify or resume:
            params = {'id': self.pk}
            if notify:
                params['steps'] = self._get_workflow_steps()
            if resume:
                params['resume'] = True
            amqp_utils.send(amqp_utils.LANDER_QUEUE, json.dumps(params))

    def _get_workflow_steps(self):
        '''build workflow steps injecting pauses where reviews are needed'''
        steps = json.loads(self.workflow.steps)
        reviews = self.review_set.filter(completed=False)
        pauses = [x.workflow_step for x in reviews]
        for step in steps:
            step_val = STEP_TO_WORKFLOW_MAPPING.get(step['name'], -1)
            if step_val in pauses:
                step['pause'] = {}
        return steps


class SourcePackageUpload(models.Model):
    class Meta:
        db_table = 'sourcepackageupload'

    sourcepackage = models.ForeignKey(SourcePackage)
    version = models.CharField(max_length=4096)

    def __unicode__(self):
        return "{} {}".format(self.sourcepackage, self.version)


class MergeProposal(models.Model):
    class Meta:
        db_table = 'mergeproposal'

    project = models.CharField(max_length=4096)
    lp_url = models.CharField(max_length=4096)
    target_branch = models.CharField(max_length=4096)
    approved_revno = models.IntegerField()
    # All branches submitted to ci-airline will result on a package, so
    # will have a debian packaging dir with the latest version.
    sourcepackage = models.ForeignKey(SourcePackage)
    version = models.CharField(max_length=4096)

    def __unicode__(self):
        return "{} {}".format(self.project, self.lp_url, self.sourcepackage,
                              self.version)

    @property
    def proposed_branch(self):
        return "lp:{}".format(self.lp_url.split("/+merge")[0])


class SubTicket(models.Model):
    class Meta:
        db_table = 'subticket'

    current_workflow_step = models.IntegerField(
        choices=_choices(SubTicketWorkflowStep), max_length=4096,
        default=000)
    status = models.IntegerField(
        choices=_choices(SubTicketWorkflowStepStatus), max_length=4096,
        default=000)
    ticket = models.ForeignKey(Ticket)
    assignee = models.EmailField(max_length=254)
    source_package_upload = models.ForeignKey(SourcePackageUpload, null=True,
                                              blank=True, default=None)
    merge_proposal = models.ForeignKey(MergeProposal, null=True, blank=True,
                                       default=None)

    def save(self, *args, **kwargs):
        if self.source_package_upload and self.merge_proposal:
            raise IntegrityError("Subtickets only accept a source package "
                                 "upload or a merge proposal at a time, you "
                                 "provided both.")
        elif not self.source_package_upload and not self.merge_proposal:
            raise IntegrityError("A subticket requires at least one source "
                                 "package upload or one merge proposal, you "
                                 "provided neither.")
        super(SubTicket, self).save(*args, **kwargs)

    def __unicode__(self):
        if self.source_package_upload:
            target = self.source_package_upload
        else:
            target = self.merge_proposal
        return "{} {}".format(self.ticket, target)


class ArtifactBase(models.Model):
    class Meta:
        abstract = True

    # 'reference' provided by the artifact manager
    reference = models.CharField(max_length=4096)
    name = models.CharField(max_length=4096)


class SubTicketArtifact(ArtifactBase):
    SUBTICKET_ARTIFACT_TYPE = (
        ('RESULTS', 'Test Results'),
        ('SPU', 'Source Package Upload'),
        ('LOGS', 'Logs'),
    )

    class Meta:
        db_table = 'subticket_artifact'

    type = models.CharField(choices=SUBTICKET_ARTIFACT_TYPE, max_length=4096)
    subticket = models.ForeignKey(SubTicket, null=True)

    def __unicode__(self):
        return "{} {}".format(self.name, self.type)


class TicketArtifact(ArtifactBase):
    TICKET_ARTIFACT_TYPE = (
        ('RESULTS', 'Test Results'),
        ('LOGS', 'Logs'),
        ('IMAGE', 'Image'),
    )

    class Meta:
        db_table = 'ticket_artifact'

    type = models.CharField(choices=TICKET_ARTIFACT_TYPE, max_length=4096)
    ticket = models.ForeignKey(Ticket, null=True)

    def __unicode__(self):
        return "{} {}".format(self.name, self.type)


class Review(models.Model):
    ticket = models.ForeignKey(Ticket)
    workflow_step = models.IntegerField(choices=_choices(TicketWorkflowStep))
    review_type = models.CharField(max_length=4096)
    completed = models.BooleanField(default=False)

    def save(self, *args, **kwargs):
        super(Review, self).save(*args, **kwargs)
        if self.ticket.review_set.filter(completed=False).count() == 0:
            # all reviews are complete, if ticket is paused we should
            # mark it "continue"
            if int(self.ticket.status) == int(TicketWorkflowStepStatus.PAUSED):
                self.ticket.status = TicketWorkflowStepStatus.CONTINUE
                self.ticket.save()

    def __unicode__(self):
        return 'ticket-{} - {}: completed({})'.format(
            self.ticket.id, self.review_type, self.completed)
