# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import mock

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase
from project.models import SourcePackage
from ticket.models import (
    MergeProposal,
    Review,
    SourcePackageUpload,
    SubTicket,
    SubTicketArtifact,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    Ticket,
    TicketArtifact,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
    Workflow,
)


def create_sourcepackage(name='my-package'):
    sourcepackage = SourcePackage()
    sourcepackage.name = name
    sourcepackage.save()
    return sourcepackage


def create_sourcepackageupload(sourcepackage='my-package', version='1.1'):
    spu = SourcePackageUpload()
    spu.sourcepackage = sourcepackage
    spu.version = version
    spu.save()
    return spu


def create_merge_proposal(project='my-project', sourcepackage='my-package',
                          lp_url='~foobar/my-project/my-branch/+merge/123456',
                          target_branch='lp:~foobar/my-project/trunk',
                          approved_revno=100, version='1.1'):
    merge_proposal = MergeProposal()
    merge_proposal.project = project
    merge_proposal.sourcepackage = sourcepackage
    merge_proposal.lp_url = lp_url
    merge_proposal.version = version
    merge_proposal.target_branch = target_branch
    merge_proposal.approved_revno = approved_revno
    merge_proposal.save()
    return merge_proposal


def create_ticket(title="This is my ticket",
                  description="this is my ticket description",
                  bug_id='12345', owner='test@example.com',
                  current_workflow_step=TicketWorkflowStep.PKG_BUILDING,
                  status=TicketWorkflowStepStatus.INPROGRESS,
                  base_image='17', added_binaries="foo,bar", series='saucy',
                  removed_binaries="baz", master_ppa='ci-engine/ppa'):
    ticket = Ticket()
    ticket.owner = owner
    ticket.title = title
    ticket.description = description
    ticket.bug_id = bug_id
    ticket.current_workflow_step = current_workflow_step
    ticket.status = status
    ticket.base_image = base_image
    ticket.added_binaries = added_binaries
    ticket.removed_binaries = removed_binaries
    ticket.series = series
    ticket.master_ppa = master_ppa
    ticket.save()
    return ticket


def create_subticket(current_workflow_step=SubTicketWorkflowStep.PKG_BUILDING,
                     ticket=None, status=(
                         SubTicketWorkflowStepStatus.PKG_BUILDING_INPROGRESS),
                     assignee='test@example.com', source_package_upload=None,
                     merge_proposal=None):
    subticket = SubTicket()
    subticket.current_workflow_step = current_workflow_step
    subticket.status = status
    subticket.ticket = ticket
    subticket.assignee = assignee
    subticket.source_package_upload = source_package_upload
    subticket.merge_proposal = merge_proposal
    subticket.save()
    return subticket


def create_ticketartifact(name="test.log", ticket=None, reference='123abc'):
    artifact = TicketArtifact()
    artifact.name = name
    artifact.ticket = ticket
    artifact.reference = reference
    artifact.save()
    return artifact


def create_subticketartifact(name="test.log", subticket=None,
                             reference='123abc'):
    artifact = SubTicketArtifact()
    artifact.name = name
    artifact.subticket = subticket
    artifact.reference = reference
    artifact.save()
    return artifact


class TicketModelsTestCase(TestCase):
    def setUp(self):
        self.sourcepackage = create_sourcepackage()
        self.spu = create_sourcepackageupload(
            sourcepackage=self.sourcepackage)
        self.mp = create_merge_proposal(sourcepackage=self.sourcepackage)
        self.ticket = create_ticket(owner='test@example.com')

    @mock.patch('ci_utils.unit_config._unit_config')
    @mock.patch('ci_utils.data_store.DataStore.ensure_swift_client')
    def test_create_container(self, mocked_esc, mocked_auth):
        # `Ticket` supports swift container creation via ci_utils
        # `DataStore.ensure_swift_client` using available 'unit_config'.
        self.ticket.create_container()
        mocked_auth.assert_called_once()
        mocked_esc.assert_called_once()

    @mock.patch('ticket.models.Ticket.create_container')
    def test_creating_a_new_ticket(self, mocked_cc):
        # Creating a NEW ticket creates its container.
        create_ticket(
            current_workflow_step=TicketWorkflowStep.NEW,
            status=TicketWorkflowStepStatus.QUEUED)
        mocked_cc.assert_called_once()

    def test_cannot_update_a_ticket_to_new(self):
        # Due to the container creation, updating a ticket *back* to
        # NEW is not allowed.
        self.ticket.current_workflow_step = TicketWorkflowStep.NEW
        with self.assertRaises(IntegrityError) as cm:
            self.ticket.save()
        self.assertEquals(
            'Cannot update a ticket back to NEW.',
            str(cm.exception))

    def test_creating_an_artifact_subticket(self):
        subticket = create_subticket(ticket=self.ticket,
                                     assignee='test@example.com',
                                     source_package_upload=self.spu)

        self.artifact_1 = create_subticketartifact(subticket=subticket)
        artifact_in_database = SubTicketArtifact.objects.exclude(
            subticket=None)
        self.assertEquals(len(artifact_in_database), 1)
        only_artifact_in_database = artifact_in_database[0]
        self.assertEquals(only_artifact_in_database, self.artifact_1)

        self.assertEquals(only_artifact_in_database.name, "test.log")

    def test_creating_an_artifact_ticket(self):
        self.artifact_2 = create_ticketartifact(ticket=self.ticket)
        artifact_in_database = TicketArtifact.objects.exclude(ticket=None)
        self.assertEquals(len(artifact_in_database), 1)
        only_artifact_in_database = artifact_in_database[0]
        self.assertEquals(only_artifact_in_database, self.artifact_2)

        self.assertEquals(only_artifact_in_database.name, "test.log")

    def test_creating_a_ticket(self):
        ticket_in_database = Ticket.objects.all()
        self.assertEquals(len(ticket_in_database), 1)
        only_ticket_in_database = ticket_in_database[0]
        self.assertEquals(only_ticket_in_database, self.ticket)

        self.assertEquals(only_ticket_in_database.title, "This is my ticket")

        self.assertEquals(only_ticket_in_database.added_binaries, "foo,bar")
        self.assertEquals(only_ticket_in_database.removed_binaries, "baz")

    def test_creating_a_subticket_with_spu(self):
        subticket = create_subticket(ticket=self.ticket,
                                     assignee='test@example.com',
                                     source_package_upload=self.spu)
        subticket_in_database = SubTicket.objects.all()
        self.assertEquals(len(subticket_in_database), 1)
        only_subticket_in_database = subticket_in_database[0]
        self.assertEquals(only_subticket_in_database, subticket)

        self.assertEquals(only_subticket_in_database.assignee,
                          "test@example.com")

    def test_creating_a_subticket_with_spu_and_mp(self):
        with self.assertRaises(IntegrityError) as cm:
            create_subticket(ticket=self.ticket, assignee='test@example.com',
                             source_package_upload=self.spu,
                             merge_proposal=self.mp)
        self.assertEquals(cm.exception.args[0], "Subtickets only accept a "
                          "source package upload or a merge proposal at a "
                          "time, you provided both.")
        subticket_in_database = SubTicket.objects.all()
        self.assertEquals(len(subticket_in_database), 0)

    def test_creating_a_subticket_with_no_spu_or_mp(self):
        with self.assertRaises(IntegrityError) as cm:
            create_subticket(ticket=self.ticket, assignee='test@example.com')
        self.assertEquals(cm.exception.args[0], "A subticket requires at "
                          "least one source package upload or one merge "
                          "proposal, you provided neither.")
        subticket_in_database = SubTicket.objects.all()
        self.assertEquals(len(subticket_in_database), 0)

    def test_creating_a_subticket_with_mp(self):
        subticket = create_subticket(ticket=self.ticket,
                                     assignee='test@example.com',
                                     merge_proposal=self.mp)
        subticket_in_database = SubTicket.objects.all()
        self.assertEquals(len(subticket_in_database), 1)
        only_subticket_in_database = subticket_in_database[0]
        self.assertEquals(only_subticket_in_database, subticket)

        self.assertEquals(only_subticket_in_database.assignee,
                          "test@example.com")

    def test_creating_a_sourcepackageupload(self):
        spu_in_database = SourcePackageUpload.objects.all()
        self.assertEquals(len(spu_in_database), 1)
        only_spu_in_database = spu_in_database[0]
        self.assertEquals(only_spu_in_database, self.spu)

        self.assertEquals(only_spu_in_database.sourcepackage.name,
                          "my-package")

    def test_creating_a_mergeproposal(self):
        mp_in_database = MergeProposal.objects.all()
        self.assertEquals(len(mp_in_database), 1)
        only_mp_in_database = mp_in_database[0]
        self.assertEquals(only_mp_in_database, self.mp)

        self.assertEquals(only_mp_in_database.project, "my-project")
        self.assertEquals(only_mp_in_database.lp_url,
                          "~foobar/my-project/my-branch/+merge/123456")
        self.assertEquals(only_mp_in_database.target_branch,
                          "lp:~foobar/my-project/trunk")
        self.assertEquals(only_mp_in_database.proposed_branch,
                          "lp:~foobar/my-project/my-branch")

    def test_added_binary_list_invalid(self):
        added_binaries = "foo bar"
        self.assertRaises(IntegrityError, create_ticket,
                          added_binaries=added_binaries)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_lander_queued(self, send):
        '''ensure a ticket that becomes queued notifies the lander'''
        self.ticket.current_workflow_step = TicketWorkflowStep.QUEUED
        self.ticket.status = TicketWorkflowStepStatus.QUEUED
        self.ticket.save()
        self.assertEqual(1, send.call_count)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_lander_continue(self, send):
        self.ticket.current_workflow_step = TicketWorkflowStep.PKG_PUBLISHING
        self.ticket.status = TicketWorkflowStepStatus.CONTINUE
        self.ticket.save()
        self.assertEqual(1, send.call_count)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_review(self, send):
        '''ensure a queued ticket with review has pause step'''
        Review.objects.create(ticket=self.ticket, review_type='packaging',
                              workflow_step=TicketWorkflowStep.PKG_PUBLISHING)
        self.ticket.current_workflow_step = TicketWorkflowStep.QUEUED
        self.ticket.status = TicketWorkflowStepStatus.QUEUED
        self.ticket.save()

        self.assertEqual(1, send.call_count)
        data = json.loads(send.call_args[0][1])
        for step in data['steps']:
            if step['name'] == 'publisher':
                self.assertTrue('pause' in step)
                return
        raise ValueError('publisher step not found in: %s' % data)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_review_continue(self, send):
        '''Ensure when reviews are marked approved the ticket continues'''
        self.ticket.current_workflow_step = TicketWorkflowStep.PKG_BUILDING
        self.ticket.status = TicketWorkflowStepStatus.PAUSED
        self.ticket.save()

        r = Review.objects.create(
            ticket=self.ticket, review_type='packaging',
            workflow_step=TicketWorkflowStep.PKG_PUBLISHING)
        self.ticket.save()
        self.assertEqual(0, send.call_count)

        # This will trigger the ticket object to be moved to CONTINUE
        r.completed = True
        r.save()
        status = Ticket.objects.get(pk=self.ticket.id).status
        self.assertEqual(TicketWorkflowStepStatus.CONTINUE.value, status)
        self.assertEqual(1, send.call_count)


class WorkflowTestCase(TestCase):
    def test_unique_name(self):
        '''make sure the workflow name is unique'''
        Workflow.objects.create(name='foo', steps='[]')
        with self.assertRaises(IntegrityError):
            Workflow.objects.create(name='foo', steps='[]')

    def test_json(self):
        '''make sure we don't allow invalid json'''
        with self.assertRaises(ValidationError):
            Workflow.objects.create(name='foo', steps='this is not json')
